import json
import logging
import os
from functools import partial

import dspy
import dspy.evaluate
import hydra
import tqdm
from dspy.datasets import HotPotQA
from dspy.dsp.utils import deduplicate
from hydra.core.hydra_config import HydraConfig
from omegaconf import DictConfig

from tapeagents.agent import Agent, Node
from tapeagents.batch import batch_main_loop
from tapeagents.core import StepMetadata, Tape
from tapeagents.dialog_tape import (
    AssistantStep,
    AssistantThought,
    DialogTape,
    UserStep,
)
from tapeagents.environment import ToolEnvironment
from tapeagents.io import save_agent, stream_yaml_tapes
from tapeagents.llm_function import LLMFunctionNode, by_node, by_step
from tapeagents.llms import LiteLLM, LLMStream
from tapeagents.optimize import OptimizationResult, optimize_demos
from tapeagents.orchestrator import main_loop
from tapeagents.renderers.camera_ready_renderer import CameraReadyRenderer
from tapeagents.renderers.pretty import PrettyRenderer
from tapeagents.studio import Studio
from tapeagents.tape_browser import TapeBrowser
from tapeagents.tool_calling import FunctionCall, ToolCall, ToolCalls, ToolResult

from .func_templates import make_answer_template, make_query_template
from .load_demos import load_agentic_rag_demos, load_rag_demos

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)
logging.getLogger("httpx").setLevel(logging.WARNING)


def make_env(n_paragraphs: int = 3) -> ToolEnvironment:
    def retrieve(query: str) -> list[str]:
        """Retrieve Wikipedia abstracts"""
        results = dspy.ColBERTv2(url="http://20.102.90.50:2017/wiki17_abstracts")(query)
        return [r["text"] for r in results[:n_paragraphs]]

    return ToolEnvironment(tools=[retrieve])


def make_llm(cfg: DictConfig) -> LiteLLM:
    parameters = {
        "temperature": 0.0,
        "max_tokens": 1024,
        "top_p": 1,
        "frequency_penalty": 0,
        "n": 1,
    }

    llm = LiteLLM(model_name=cfg.llm_name, parameters=parameters, use_cache=cfg.llm_cache)
    return llm


def make_rag_agent(cfg: DictConfig) -> Agent:
    class RetrieveNode(Node):
        def generate_steps(self, agent, tape: Tape, llm_stream: LLMStream):
            query = tape.steps[-1].content
            tc = ToolCall(function=FunctionCall(name="retrieve", arguments={"query": query}))
            yield ToolCalls(tool_calls=[tc])

    agent = Agent.create(
        llms=make_llm(cfg),
        templates={"rag": make_answer_template()},
        nodes=[RetrieveNode(), LLMFunctionNode(template_name="rag", input_refs=[-1, -3])],
    )
    if cfg.load_demos:
        partial_demos, demos = load_rag_demos()
        agent.templates["rag"].demos = demos
        agent.templates["rag"].partial_demos = partial_demos
    return agent


def make_agentic_rag_agent(cfg: DictConfig) -> Agent:
    templates = {
        "answer": make_answer_template(),
    }
    for i in range(cfg.agentic_rag.max_hops):
        templates[f"query{i}"] = make_query_template()

    class Deduplicate(Node):
        def generate_steps(self, agent, tape: Tape, llm_stream: LLMStream):
            contexts = []
            for step in tape:
                if isinstance(step, ToolResult):
                    contexts.extend(step.content)
            yield AssistantThought(content=deduplicate(contexts))

    nodes = []
    for i in range(cfg.agentic_rag.max_hops):
        context_ref = by_step(ToolResult) if i > 0 else AssistantThought(content=[])
        nodes.append(
            LLMFunctionNode(
                name=f"query{i}",
                template_name=f"query{i}",
                input_refs=[context_ref, by_step(UserStep)],
            )
        )
        nodes.append(Deduplicate(name=f"deduplicate{i}"))
    nodes.append(LLMFunctionNode(template_name="answer", input_refs=[by_node(nodes[-1]), by_step(UserStep)]))

    agent = Agent.create(llms=make_llm(cfg), templates=templates, nodes=nodes)

    if cfg.load_demos:
        all_demos = load_agentic_rag_demos()
        for template_name, (partial_demos, demos) in all_demos.items():
            agent.templates[template_name].demos = demos
            agent.templates[template_name].partial_demos = partial_demos

    return agent


def run_agent(agent: Agent, dataset: list[dspy.primitives.Example], cfg: DictConfig) -> tuple[list[Tape], list[Tape]]:
    env = make_env(cfg.optimize.n_paragraphs)
    start_tapes = make_start_tapes(dataset)
    final_tapes = list(tqdm.tqdm(batch_main_loop(agent, start_tapes, env)))
    return final_tapes


def optimize_agent(agent: Agent, cfg: DictConfig) -> Agent:
    # Step 1: Run agent on the training set
    dataset = get_dataset(cfg)
    final_tapes = run_agent(agent, dataset.train, cfg)
    # Step 2: filter out good tapes
    good_tapes = [tape for tape in final_tapes if is_good_tape(tape)]
    bad_tapes = [tape for tape in final_tapes if tape not in good_tapes]
    logger.info(f"{len(good_tapes)} good tapes out of {len(final_tapes)}")
    # Save all tapes for observability
    with stream_yaml_tapes("good_training_tapes.yaml") as saver:
        for tape in good_tapes:
            saver.save(tape)
    with stream_yaml_tapes("bad_training_tapes.yaml") as saver:
        for tape in bad_tapes:
            saver.save(tape)
    # Step 3: Optimize agent from the good tapes
    run_agent_with_val_dataset = partial(run_agent, dataset=dataset.dev, cfg=cfg)
    best_optimization = optimize_demos(
        agent=agent,
        good_tapes=good_tapes,
        n_demos=cfg.optimize.max_n_demos,
        n_iterations=cfg.optimize.max_optimize_tries,
        seed=cfg.seed,
        metric_fn=compute_weighted_accuracy,
        run_agent_fn=run_agent_with_val_dataset,
        post_run_agent_fn=save_optimization_result,
    )
    save_optimization_result(result=best_optimization, run_name="best")
    return best_optimization.agent


def save_optimization_result(result: OptimizationResult, run_name: str = "optimization") -> None:
    filename = f"agent_{f'{run_name}_' if run_name else ''}{result.id}.yaml"
    save_agent(agent=result.agent, filename=filename)


def compute_weighted_accuracy(
    tapes: list[Tape], run_name: str = "", w_retrieval: float = 0.5, w_answer: float = 0.5
) -> float:
    """Compute the weighted average of the retrival and answer accuracy"""
    retrieval_accuracy = compute_retrieval_accuracy(tapes)
    answer_accuracy = compute_answer_exact_match(tapes)
    mean_accuracy = retrieval_accuracy * w_retrieval + answer_accuracy * w_answer
    metrics = {
        "mean_accuracy": mean_accuracy,
        "retrieval_accuracy": retrieval_accuracy,
        "answer_accuracy": answer_accuracy,
    }
    # Log metrics and save
    logger.info(metrics)
    metrics_save_path = f"metrics_{len(tapes)}{f'_{run_name}' if run_name else ''}.json"
    save_metrics(metrics, metrics_save_path)

    return mean_accuracy


def save_metrics(metrics: dict, path: str) -> None:
    with open(path, "w") as f:
        json.dump(
            metrics,
            f,
            indent=2,
        )


def make_agent(cfg: DictConfig) -> Agent:
    agent = make_rag_agent(cfg) if cfg.agent == "rag" else make_agentic_rag_agent(cfg)
    if cfg.optimize.do:
        agent = optimize_agent(agent, cfg)
    return agent


def is_good_tape(tape):
    if len(tape.steps) < 3:  # Error in the tape
        logger.error(tape.metadata.error)
        return False
    pred = dspy.primitives.Example({"answer": str(tape.steps[-1].content).strip(), "context": tape.steps[-3].content})
    expected_answer = dspy.primitives.Example({"answer": tape.steps[0].metadata.other["answer"]})
    if not dspy.evaluate.answer_exact_match(expected_answer, pred):
        tape.metadata.result["reason"] = "bad answer"
        return False
    if not dspy.evaluate.answer_passage_match(expected_answer, pred):
        tape.metadata.result["reason"] = "answer not in context"
        return False
    queries = [tape.steps[0].content]
    queries += [step.tool_calls[0].function.arguments["query"] for step in tape if isinstance(step, ToolCalls)]
    if max([len(q) for q in queries]) > 200:
        tape.metadata.result["reason"] = "long query"
        return False
    if any(
        dspy.evaluate.answer_exact_match_str(queries[idx], queries[:idx], frac=0.8) for idx in range(2, len(queries))
    ):
        tape.metadata.result["reason"] = "repeated query"
        return False
    tape.metadata.result["reason"] = "good tape"
    return True


def compute_retrieval_accuracy(tapes: list[Tape]):
    n_correct = 0
    for tape in tapes:
        gold_titles = set(map(dspy.evaluate.normalize_text, tape.steps[0].metadata.other["gold_titles"]))
        if len(tape.steps) < 3:  # Error in the tape
            tape.metadata.result["retrieval_accurate"] = False
            continue
        context_step = tape.steps[-3]
        found_titles = [c.split(" | ")[0] for c in context_step.content]
        found_titles = set(map(dspy.evaluate.normalize_text, found_titles))
        ok = gold_titles.issubset(found_titles)
        tape.metadata.result["retrieval_accurate"] = ok
        n_correct += int(ok)
    return n_correct / len(tapes)


def compute_answer_exact_match(tapes: list[Tape]):
    n_correct = 0
    for tape in tapes:
        expected_answer = tape.steps[0].metadata.other["answer"]
        answer_step = tape.steps[-1]
        answer_text = str(answer_step.content).strip()
        if isinstance(answer_step, AssistantStep):
            ok = dspy.evaluate.answer_exact_match(
                dspy.primitives.Example({"answer": expected_answer}), dspy.primitives.Example({"answer": answer_text})
            )
            tape.metadata.result["answer_accurate"] = ok
            n_correct += int(ok)
    return n_correct / len(tapes)


_dataset = None


def get_dataset(cfg: DictConfig) -> HotPotQA:
    logger.info("Loading data ...")
    global _dataset
    if _dataset is None:
        _dataset = HotPotQA(
            train_seed=1,
            train_size=cfg.dataset.train_size,
            eval_seed=2023,
            dev_size=cfg.dataset.dev_size,
            test_size=cfg.dataset.test_size,
        )
    logger.info("Data loaded")
    return _dataset


def make_start_tapes(dataset: list[dspy.primitives.Example]) -> list[DialogTape]:
    # Note: gold_titles are not present on the training set
    return [
        DialogTape(
            steps=[
                UserStep(
                    metadata=StepMetadata(
                        other={"answer": example["answer"], "gold_titles": example.get("gold_titles", set())}
                    ),
                    content=example["question"],
                )
            ],
        )
        for example in dataset
    ]


def batch_run_and_save(
    agent: Agent, env: ToolEnvironment, dataset: list[dspy.primitives.Example], save_tapes_path: str
):
    start_tapes = make_start_tapes(dataset)
    final_tapes = []
    with stream_yaml_tapes(save_tapes_path) as saver:
        for tape in tqdm.tqdm(batch_main_loop(agent, start_tapes, env)):
            final_tapes.append(tape)
            saver.save(tape)
    return final_tapes


def run(cfg: DictConfig):
    agent = make_agent(cfg)
    env = make_env()
    start_tape = DialogTape(steps=[UserStep(content=cfg.question)])
    final_tape = main_loop(agent, start_tape, env).get_final_tape()
    final_tape_str = final_tape.model_dump_json(indent=2)
    with open("tape.json", "w") as f:
        print(final_tape_str, file=f)
    print(final_tape_str)


def studio(cfg: DictConfig):
    agent = make_agent(cfg)
    env = make_env()
    start_tape = DialogTape(steps=[UserStep(content=cfg.question)])
    Studio(agent, start_tape, PrettyRenderer(), env).launch()


def evaluate(cfg: DictConfig):
    agent = make_agent(cfg)
    env = make_env()
    dataset = get_dataset(cfg)
    tapes_save_path = f"test_tapes_{cfg.dataset.test_size}.yaml"
    final_tapes = batch_run_and_save(agent, env, dataset.test, tapes_save_path)
    compute_weighted_accuracy(final_tapes, run_name="test")


def browse():
    run_dir = HydraConfig.get().runtime.output_dir
    browser = TapeBrowser(DialogTape, run_dir, CameraReadyRenderer())
    browser.launch()


@hydra.main(version_base=None, config_path="../../conf", config_name="optimize_hotpotqa")
def main(cfg: DictConfig):
    print(f"Running in {os.getcwd()}")
    match cfg.target:
        case "run":
            run(cfg)
        case "studio":
            studio(cfg)
        case "evaluate":
            evaluate(cfg)
        case "browse":
            browse()
        case _:
            raise ValueError(f"Unknown target {cfg.target}")


if __name__ == "__main__":
    main()



import gzip
import logging
import os
import shutil
from pathlib import Path

import yaml
from omegaconf import DictConfig

from .evaluate import main as evaluate

logging.basicConfig(level=logging.INFO)

test_path = Path("tests/examples/res/gaia_agent/")

run_dir = Path("/tmp/gaiatest")
if run_dir.exists():
    print(f"Tmp dir exists {run_dir}, removing")
    shutil.rmtree(run_dir)

run_dir.mkdir(parents=True)
cache_dir = run_dir / "cache"
cache_dir.mkdir(parents=True, exist_ok=True)
os.environ["TAPEAGENTS_CACHE_DIR"] = str(cache_dir)
os.environ["TAPEAGENTS_SQLITE_DB"] = str(run_dir / "tapedata.sqlite")
(run_dir / ".hydra").mkdir(parents=True, exist_ok=True)
shutil.copyfile(test_path / "config.yaml", run_dir / ".hydra" / "config.yaml")
print("Tmp dir ready:", os.listdir(run_dir))

with open(test_path / "config.yaml") as f:
    cfg = DictConfig(yaml.safe_load(f))
cfg.exp_path = str(run_dir)
cfg.environment.tools[2].mock = False
evaluate(cfg)

print("Done, now copy the tapes, cache and db")
shutil.copytree(run_dir / "tapes", test_path / "tapes", dirs_exist_ok=True)
shutil.copytree(run_dir / "cache", test_path / "cache", dirs_exist_ok=True)
with open(run_dir / "tapedata.sqlite", "rb") as f_in:
    with gzip.open(test_path / "tapedata.sqlite.gz", "wb") as f_out:
        shutil.copyfileobj(f_in, f_out)
print("Test data updated, now run the tests with `make test`")



### copy-paste of llmd2.core.error ####

from abc import ABC, abstractmethod
from io import StringIO
from typing import Any, Literal, Union

from pydantic import Field, JsonValue
from typing_extensions import Self

from tapeagents.core import Action, Error, Step

from .types import FunctionName, ParameterName

# from .steps import FormFillerStep


class FormFillerStateError(ABC, Action, Error):
    """
    A base class for the form filler state errors.
    Errors are used to indicate that a step failed.
    They are not raised as exceptions but rather returned as a special step.
    They can be converted to exceptions using the :meth:`to_exc()` method.
    """

    kind: Literal[  # type: ignore
        "unknown_function_error",
        "unknown_function_schema_error",
        "invalid_function_schema_error",
        "unknown_function_parameter_error",
        "invalid_function_parameter_value_error",
        "valid_function_parameter_value_error",
        "invalid_function_parameter_skip_error",
        "valid_function_parameter_skip_error",
        "invalid_function_parameters_error",
        "function_call_error",
        "invalid_function_return_value_error",
        "unknown_error",
    ]
    # we use generic type here to avoid circular imports
    cause: Any = None

    @abstractmethod
    def to_exc(self: Self, turn: None | int = None, step: None | int = None) -> BaseException:
        """
        Convert the error to an exception that can be raised.
        """
        ...

    def add_cause(self, cause: Step) -> Self:
        return self.model_copy(update={"cause": cause})

    def __bool__(self: Self) -> bool:
        """
        Errors are always falsey.
        """
        return False


class ValueErrorWithContext(ValueError):
    def __init__(self, message: str, turn: int | None = None, step: int | None = None):
        super().__init__(message)
        self.turn = turn
        self.step = step


class UnknownFunctionError(FormFillerStateError):
    """
    An error that indicates that the function is unknown.
    """

    kind: Literal["unknown_function_error"] = "unknown_function_error"
    function: FunctionName
    message: str = ""

    def to_exc(self: Self, turn: None | int = None, step: None | int = None) -> BaseException:
        msg = StringIO()
        if turn is not None:
            msg.write(f"Turn {turn}: ")
        if step is not None:
            msg.write(f"Step {step}: ")
        msg.write(f"Unknown function {self.function!r}")
        if self.message:
            msg.write(f": {self.message}")
        return ValueErrorWithContext(msg.getvalue(), turn=turn, step=step)


class UnknownFunctionSchemaError(FormFillerStateError):
    """
    An error that indicates that the function schema is unknown.
    """

    kind: Literal["unknown_function_schema_error"] = "unknown_function_schema_error"
    function: FunctionName
    message: str = ""

    def to_exc(self: Self, turn: None | int = None, step: None | int = None) -> BaseException:
        msg = StringIO()
        if turn is not None:
            msg.write(f"Turn {turn}: ")
        if step is not None:
            msg.write(f"Step {step}: ")
        msg.write(f"Unknown schema for function {self.function!r}")
        if self.message:
            msg.write(f": {self.message}")
        return ValueErrorWithContext(msg.getvalue(), turn=turn, step=step)


class InvalidFunctionSchemaError(FormFillerStateError):
    """
    An error that indicates that the function schema is invalid.
    """

    kind: Literal["invalid_function_schema_error"] = "invalid_function_schema_error"
    function: FunctionName
    # we use Any here to avoid circular imports
    function_schema: Any
    message: str = ""

    def to_exc(self: Self, turn: None | int = None, step: None | int = None) -> BaseException:
        msg = StringIO()
        if turn is not None:
            msg.write(f"Turn {turn}: ")
        if step is not None:
            msg.write(f"Step {step}: ")
        msg.write(f"Invalid schema {self.function_schema!r} for function {self.function!r}")
        if self.message:
            msg.write(f": {self.message}")
        return ValueErrorWithContext(msg.getvalue(), turn=turn, step=step)


class UnknownFunctionParameterError(FormFillerStateError):
    """
    An error that indicates that the function parameter is unknown.
    """

    kind: Literal["unknown_function_parameter_error"] = "unknown_function_parameter_error"
    function: FunctionName
    parameter: ParameterName
    message: str = ""

    def to_exc(self: Self, turn: None | int = None, step: None | int = None) -> BaseException:
        msg = StringIO()
        if turn is not None:
            msg.write(f"Turn {turn}: ")
        if step is not None:
            msg.write(f"Step {step}: ")
        msg.write(f"Unknown parameter {self.parameter!r} of function {self.function!r}")
        if self.message:
            msg.write(f": {self.message}")
        return ValueErrorWithContext(msg.getvalue(), turn=turn, step=step)


class InvalidFunctionParameterValueError(FormFillerStateError):
    """
    An error that indicates that the value of a function parameter is invalid.
    """

    kind: Literal["invalid_function_parameter_value_error"] = "invalid_function_parameter_value_error"
    function: FunctionName
    parameter: ParameterName
    parameter_value: JsonValue
    message: str = ""

    def to_exc(self: Self, turn: None | int = None, step: None | int = None) -> BaseException:
        msg = StringIO()
        if turn is not None:
            msg.write(f"Turn {turn}: ")
        if step is not None:
            msg.write(f"Step {step}: ")
        msg.write(
            f"Invalid value {self.parameter_value!r} for parameter {self.parameter!r} of function {self.function!r}"
        )
        if self.message:
            msg.write(f": {self.message}")
        return ValueErrorWithContext(msg.getvalue(), turn=turn, step=step)


class ValidFunctionParameterValueError(FormFillerStateError):
    """
    An error that indicates that the value of a function parameter is valid but should have been invalid.
    """

    kind: Literal["valid_function_parameter_value_error"] = "valid_function_parameter_value_error"
    function: FunctionName
    parameter: ParameterName
    parameter_value: JsonValue
    message: str = ""

    def to_exc(self: Self, turn: None | int = None, step: None | int = None) -> BaseException:
        msg = StringIO()
        if turn is not None:
            msg.write(f"Turn {turn}: ")
        if step is not None:
            msg.write(f"Step {step}: ")
        msg.write(
            f"Valid value {self.parameter_value!r} for parameter {self.parameter!r} of function {self.function!r}"
        )
        if self.message:
            msg.write(f": {self.message}")
        return ValueErrorWithContext(msg.getvalue(), turn=turn, step=step)


class InvalidFunctionParameterSkipError(FormFillerStateError):
    """
    An error that indicates that the parameter was skipped but was not allowed to be skipped.
    """

    kind: Literal["invalid_function_parameter_skip_error"] = "invalid_function_parameter_skip_error"
    function: FunctionName
    parameter: ParameterName
    message: str = ""

    def to_exc(self: Self, turn: None | int = None, step: None | int = None) -> BaseException:
        msg = StringIO()
        if turn is not None:
            msg.write(f"Turn {turn}: ")
        if step is not None:
            msg.write(f"Step {step}: ")
        msg.write(f"Unexpected skip for parameter {self.parameter!r} of function {self.function!r}")
        if self.message:
            msg.write(f": {self.message}")
        return ValueErrorWithContext(msg.getvalue(), turn=turn, step=step)


class ValidFunctionParameterSkipError(FormFillerStateError):
    """
    An error that indicates that the parameter could have been skipped but was not skipped.
    """

    kind: Literal["valid_function_parameter_skip_error"] = "valid_function_parameter_skip_error"
    function: FunctionName
    parameter: ParameterName
    message: str = ""

    def to_exc(self: Self, turn: None | int = None, step: None | int = None) -> BaseException:
        msg = StringIO()
        if turn is not None:
            msg.write(f"Turn {turn}: ")
        if step is not None:
            msg.write(f"Step {step}: ")
        msg.write(f"Valid skip for parameter {self.parameter!r} of function {self.function!r}")
        if self.message:
            msg.write(f": {self.message}")
        return ValueErrorWithContext(msg.getvalue(), turn=turn, step=step)


class InvalidFunctionParametersError(FormFillerStateError):
    """
    An error that indicates that the values of the function parameters are invalid as a whole.
    """

    kind: Literal["invalid_function_parameters_error"] = "invalid_function_parameters_error"
    function: FunctionName
    parameters: dict[ParameterName, JsonValue]
    missing_parameters: list[ParameterName] = Field(default=[])
    message: str = ""

    def to_exc(self: Self, turn: None | int = None, step: None | int = None) -> BaseException:
        msg = StringIO()
        if turn is not None:
            msg.write(f"Turn {turn}: ")
        if step is not None:
            msg.write(f"Step {step}: ")
        msg.write(f"Invalid values {self.parameters!r} for parameters of function {self.function!r}")
        if self.message:
            msg.write(f": {self.message}")
        return ValueErrorWithContext(msg.getvalue(), turn=turn, step=step)


class FunctionCallError(FormFillerStateError):
    """
    An error that indicates that the call to a function failed.
    The function was successfullly called but returned an empty result or an error.
    """

    kind: Literal["function_call_error"] = "function_call_error"
    function: FunctionName
    message: str = ""

    def to_exc(self: Self, turn: None | int = None, step: None | int = None) -> BaseException:
        msg = StringIO()
        if turn is not None:
            msg.write(f"Turn {turn}: ")
        if step is not None:
            msg.write(f"Step {step}: ")
        msg.write(f"Call to function {self.function!r} failed")
        if self.message:
            msg.write(f": {self.message}")
        return ValueErrorWithContext(msg.getvalue(), turn=turn, step=step)


class InvalidFunctionReturnValueError(FormFillerStateError):
    """
    An error that indicates that the return value of a function is invalid given its schema.
    """

    kind: Literal["invalid_function_return_value_error"] = "invalid_function_return_value_error"
    function: FunctionName
    return_value: JsonValue
    message: str = ""

    def to_exc(self: Self, turn: None | int = None, step: None | int = None) -> BaseException:
        msg = StringIO()
        if turn is not None:
            msg.write(f"Turn {turn}: ")
        if step is not None:
            msg.write(f"Step {step}: ")
        msg.write(f"Return value {self.return_value!r} of function {self.function!r} is invalid")
        if self.message:
            msg.write(f": {self.message}")
        return ValueErrorWithContext(msg.getvalue(), turn=turn, step=step)


class EnvironmentActionError(FormFillerStateError):
    """
    An error that indicates that an error occurred during action execution the environment.
    """

    kind: Literal["environment_action_error"] = "environment_action_error"
    message: str = ""

    def to_exc(self: Self, turn: None | int = None, step: None | int = None) -> BaseException:
        return RuntimeError(f"Environment action error: {self.message}")


class UnknownError(FormFillerStateError):
    """
    An error that indicates that an unknown error occurred.
    """

    kind: Literal["unknown_error"] = "unknown_error"
    message: str = ""

    def to_exc(self: Self, turn: None | int = None, step: None | int = None) -> BaseException:
        msg = StringIO()
        if turn is not None:
            msg.write(f"Turn {turn}: ")
        if step is not None:
            msg.write(f"Step {step}: ")
        msg.write("Unknown error")
        if self.message:
            msg.write(f": {self.message}")
        return ValueErrorWithContext(self.message, turn=turn, step=step)


_FormFillerStateError = Union[
    UnknownFunctionError,
    UnknownFunctionSchemaError,
    InvalidFunctionSchemaError,
    UnknownFunctionParameterError,
    InvalidFunctionParameterValueError,
    ValidFunctionParameterValueError,
    InvalidFunctionParameterSkipError,
    ValidFunctionParameterSkipError,
    InvalidFunctionParametersError,
    FunctionCallError,
    InvalidFunctionReturnValueError,
    EnvironmentActionError,
    UnknownError,
]



# read json files from a folder, create new json with the same name that contains all the content

import json
import os
import sys


def gather_jsons(folder: str):
    all_jsons = []
    for root, _, files in os.walk(folder):
        for file in files:
            if file.endswith(".json"):
                with open(os.path.join(root, file)) as f:
                    all_jsons.append(json.load(f))

    dst_dir = f"{folder}/all"
    os.makedirs(dst_dir, exist_ok=True)
    dst_name = f"{dst_dir}/tapes.json"
    with open(dst_name, "w") as f:
        json.dump(all_jsons, f, indent=4)


gather_jsons(sys.argv[1])



import hydra
from omegaconf import DictConfig

from tapeagents.dialog_tape import DialogTape, UserStep
from tapeagents.orchestrator import get_agent_and_env_from_config, main_loop


def do_loop(agent, start_tape, env):
    for event in main_loop(agent, start_tape, env):
        if ae := event.agent_event:
            if ae.step:
                print(ae.step.model_dump_json(indent=2))
            if ae.final_tape:
                final_tape = ae.final_tape
        if event.observation:
            print(event.observation.model_dump_json(indent=2))
    return final_tape


@hydra.main(
    version_base=None,
    config_path="../conf",
    config_name="use_memory",
)
def main(cfg: DictConfig):
    agent, env = get_agent_and_env_from_config(cfg)

    print("Run the agent!")
    do_loop(
        agent, DialogTape() + [UserStep(content="Perform the calculation 2 + 2. Perform the calculation 3 + 3.")], env
    )
    do_loop(
        agent, DialogTape() + [UserStep(content="Perform the calculation 2 + 2. Perform the calculation 5 + 2.")], env
    )


if __name__ == "__main__":
    main()



import datetime
import getpass
import logging
import os

import hydra
from omegaconf import DictConfig

from examples.gaia_agent.steps import GaiaQuestion, GaiaTape
from tapeagents.core import Action, Observation, Step
from tapeagents.dialog_tape import UserStep
from tapeagents.io import save_json_tape, save_tape_images
from tapeagents.orchestrator import get_agent_and_env_from_config, main_loop
from tapeagents.renderers import to_pretty_str
from tapeagents.steps import ReasoningThought
from tapeagents.tools.container_executor import init_code_sandbox

logger = logging.getLogger(__name__)


@hydra.main(
    version_base=None,
    config_path="../../../conf",
    config_name="gaia_demo",
)
def main(cfg: DictConfig) -> None:
    playwright_dir = ".pw-browsers"
    assert os.path.exists(playwright_dir), f"playwright browsers directory not found: {playwright_dir}"
    os.environ["PLAYWRIGHT_BROWSERS_PATH"] = playwright_dir
    os.environ["TAPEAGENTS_SQLITE_DB"] = os.path.join(cfg.exp_path, "tapedata.sqlite")
    if not os.environ.get("OPENAI_API_KEY"):
        oai_key = getpass.getpass(
            "Please enter your OpenAI API key (you can get one on https://platform.openai.com/settings/organization/api-keys) and press Enter to continue: "
        )
        os.environ["OPENAI_API_KEY"] = oai_key
    if not os.environ.get("SERPER_API_KEY"):
        serper_key = getpass.getpass(
            "Please enter your Serper API key (you can get one on https://serper.dev/api-key) and press Enter to continue: "
        )
        os.environ["SERPER_API_KEY"] = serper_key
    tapes_dir = os.path.join(cfg.exp_path, "tapes")
    os.makedirs(tapes_dir, exist_ok=True)
    images_dir = os.path.join(cfg.exp_path, "attachments", "images")
    os.makedirs(images_dir, exist_ok=True)
    init_code_sandbox(cfg.exp_path)
    agent, env = get_agent_and_env_from_config(cfg)
    chat = [t for t in env.tools if t.__class__.__name__ == "Browser"][0]._env.chat
    chat.add_message(role="assistant", msg="TapeAgent Ready")
    today_date_str = datetime.datetime.now().strftime("%Y-%m-%d")
    tape = None
    chat.wait_for_user_message()
    content = chat.messages[-1]["message"]
    while content.lower() != "stop":
        if content.lower() == "reset":
            tape = None
            chat.add_message(role="assistant", msg="Reset conversation, you can ask a new question now.")
            chat.wait_for_user_message()
            content = chat.messages[-1]["message"]
        if tape is None:
            tape = GaiaTape(steps=[GaiaQuestion(content=f"Today is {today_date_str}.\n{content}")])
        else:
            # continue the conversation, replace the last answer step with a reasoning step
            tape.steps[-1] = ReasoningThought(reasoning=tape.steps[-1].long_answer)
            tape = tape.append(UserStep(content=content))
        chat.add_message(role="assistant", msg="Thinking...")
        try:
            for event in main_loop(agent, tape, env, max_loops=50):
                if partial_tape := (event.agent_tape or event.env_tape):
                    tape = partial_tape
                if event.agent_event and event.agent_event.step:
                    step = event.agent_event.step
                    if step.kind in ["set_next_node"]:
                        continue
                    msg = render_step(step)
                    if msg:
                        chat.add_message(role="assistant", msg=msg)
                elif event.observation:
                    step = event.observation
                    msg = render_step(step)
                    if msg:
                        chat.add_message(role="assistant" if step.kind == "page_observation" else "user", msg=msg)
        except Exception as e:
            tape.metadata.error = str(e)
            logger.exception(f"Failed to solve task: {e}")
            chat.add_message(role="assistant", msg=f"Failed to solve task: {e}")
        chat.wait_for_user_message()
        content = chat.messages[-1]["message"]
    env.close()
    save_json_tape(tape, tapes_dir, "demo1")
    save_tape_images(tape, images_dir)
    logger.info(f"Saved to {tapes_dir}")


def render_step(step: Step) -> str:
    logger.info(step.short_view() if isinstance(step, Observation) else step.llm_view())
    msg = ""
    if step.kind == "llm_output_parsing_failure_action":
        msg = "LLM response error, retry"
    elif step.kind == "plan_thought":
        msg = f"Plan:\n{to_pretty_str(step.plan)}"
    elif step.kind == "page_observation":
        msg = "Reading web page..."
    elif step.kind == "facts_survey_thought":
        msg = f"Given facts:\n{to_pretty_str(step.given_facts)}"
        if len(step.facts_to_lookup) > 0:
            msg += f"\nFacts to look up:\n{to_pretty_str(step.facts_to_lookup)}"
    elif step.kind == "reasoning_thought":
        msg = step.reasoning
    elif step.kind == "reading_result_thought":
        msg = f'{step.fact_description}\nSupporting quote: "{step.quote_with_fact}"'
    elif step.kind == "gaia_answer_action":
        if step.success:
            msg = f"Answer: {step.long_answer}"
        else:
            msg = f"No answer found:\n{step.overview}"
    elif step.kind in ["python_code_action", "search_action", "watch_video_action"]:
        msg = to_pretty_str(step.llm_dict())
    elif step and step.kind == "scroll_action":
        msg = "Scrolling..."
    elif isinstance(step, Action):
        msg = "Interacting with the browser..."
    elif step and step.kind == "search_results_observation":
        msg = step.error or to_pretty_str([f"[{r['url'][:30]}...]{r['title'][:60]}..." for r in step.serp])
    elif step and step.kind == "code_execution_result":
        msg = f"Code execution result: {step.result.output or ''}"
    return msg


if __name__ == "__main__":
    main()



"""
Entrypoint for streamlit, see https://docs.streamlit.io/
"""

import asyncio
import logging
import os
import traceback
from datetime import datetime
from pathlib import Path

import streamlit as st
from hydra import compose, initialize

from tapeagents.core import Step, Tape
from tapeagents.dialog_tape import AssistantAnswer, UserStep
from tapeagents.io import save_json_tape, save_tape_images
from tapeagents.orchestrator import get_agent_and_env_from_config, main_loop
from tapeagents.steps import ReasoningThought
from tapeagents.tools.computer.steps import MouseHoverAction

# Set up all loggers to print to stdout
logging.basicConfig(
    level=logging.INFO,
    force=True,  # This overwrites any existing logger configurations
    format="%(asctime)s - %(levelname)s - %(name)s:%(lineno)d - %(funcName)s - %(message)s",
    handlers=[
        logging.StreamHandler()  # Explicitly add stdout handler
    ],
)

CONFIG_DIR = Path(".streamlit_config")
API_KEY_FILE = CONFIG_DIR / "api_key"
STREAMLIT_STYLE = """
<style>
    /* Highlight the stop button in red */
    button[kind=header] {
        background-color: rgb(255, 75, 75);
        border: 1px solid rgb(255, 75, 75);
        color: rgb(255, 255, 255);
    }
    button[kind=header]:hover {
        background-color: rgb(255, 51, 51);
    }
     /* Hide the streamlit deploy button */
    .stAppDeployButton {
        visibility: hidden;
    }
    h1 {
        font-size: 2em !important;
    }
</style>
"""

INTERRUPT_TEXT = "(user stopped or interrupted and wrote the following)"
INTERRUPT_TOOL_ERROR = "human stopped or interrupted tool execution"


def setup_state(cfg):
    if "api_key" not in st.session_state:
        # Try to load API key from file first, then environment
        st.session_state.api_key = load_from_storage("api_key") or os.getenv("OPENAI_API_KEY", "")
    else:
        os.environ["OPENAI_API_KEY"] = st.session_state.api_key
    if "serper_api_key" not in st.session_state:
        st.session_state.serper_api_key = load_from_storage("serper_api_key") or os.getenv("SERPER_API_KEY", "")
    else:
        os.environ["SERPER_API_KEY"] = st.session_state.serper_api_key
    if "grounding_api_key" not in st.session_state:
        st.session_state.grounding_api_key = load_from_storage("grounding_api_key") or os.getenv(
            "GROUNDING_API_KEY", ""
        )
    if st.session_state.grounding_api_key:
        os.environ["GROUNDING_API_KEY"] = st.session_state.grounding_api_key
    if "tape" not in st.session_state:
        st.session_state.tape = None
    if "env" not in st.session_state:
        os.environ["TAPEAGENTS_SQLITE_DB"] = os.path.join(cfg.exp_path, "tapedata.sqlite")
        agent, env = get_agent_and_env_from_config(cfg)
        st.session_state.env = env
        st.session_state.agent = agent
    if st.session_state.tape is None:
        st.session_state.tapes_dir = os.path.join(cfg.exp_path, "tapes")
        st.session_state.tape_name = f"chat_{datetime.now().strftime('%Y%m%d_%H%M%S')}"
        st.session_state.images_dir = os.path.join(cfg.exp_path, "attachments", "images")
        st.session_state.messages = [
            {"role": "assistant", "content": "Hi, TapeAgents Operator here! How can I help you today?"}
        ]
        if not st.session_state.api_key:
            st.session_state.messages.append(
                {
                    "role": "assistant",
                    "content": "Please enter your OpenAI API key in the sidebar to start the conversation.",
                }
            )
        if not st.session_state.serper_api_key:
            st.session_state.messages.append(
                {
                    "role": "assistant",
                    "content": "Please enter your Serper search API key in the sidebar to start the conversation.",
                }
            )


async def main(cfg):
    setup_state(cfg)

    st.markdown(STREAMLIT_STYLE, unsafe_allow_html=True)

    with st.sidebar:
        st.text_input(
            "Openai API Key",
            type="password",
            key="api_key",
            on_change=lambda: save_to_storage("api_key", st.session_state.api_key),
        )
        st.text_input(
            "Serper API Key",
            type="password",
            key="serper_api_key",
            on_change=lambda: save_to_storage("serper_api_key", st.session_state.serper_api_key),
        )
        st.text_input(
            "Grounding API Key",
            type="password",
            key="grounding_api_key",
            on_change=lambda: save_to_storage("grounding_api_key", st.session_state.grounding_api_key),
        )
        if st.button("Reset Conversation"):
            st.session_state.tape = None
            st.session_state.env.reset()
            st.rerun()

    today_date_str = datetime.now().strftime("%Y-%m-%d")
    if st.session_state.tape is None:
        initial_obs = st.session_state.env.step(MouseHoverAction(element_description="center of the screen"))
        st.session_state.tape = Tape(steps=[initial_obs, UserStep(content=f"Today is {today_date_str}")])
    else:
        if isinstance(st.session_state.tape.steps[-1], AssistantAnswer):
            st.session_state.tape.steps[-1] = ReasoningThought(reasoning=st.session_state.tape.steps[-1].answer)

    for message in st.session_state.messages:
        with st.chat_message(message["role"]):
            msg_type = message.get("type", "write")
            if msg_type == "code":
                st.code(message["content"], language="python")
            elif msg_type == "html":
                st.html(message["content"])
            elif msg_type == "markdown":
                st.markdown(message["content"])
            else:
                st.write(message["content"])

    if prompt := st.chat_input():
        with st.chat_message("user"):
            st.write(prompt)
        st.session_state.messages.append({"role": "user", "content": prompt})

        last_status = st.status("Thinking...")
        st.session_state.tape = st.session_state.tape.append(UserStep(content=prompt))

        try:
            for event in main_loop(st.session_state.agent, st.session_state.tape, st.session_state.env, max_loops=50):
                msg = None
                icon = ""
                if partial_tape := (event.agent_tape or event.env_tape):
                    st.session_state.tape = partial_tape
                    save_json_tape(partial_tape, st.session_state.tapes_dir, st.session_state.tape_name)
                    save_tape_images(partial_tape, st.session_state.images_dir)
                if event.agent_event and event.agent_event.step:
                    step = event.agent_event.step
                    if step.kind in ["set_next_node"]:
                        continue
                    msg, msg_type, icon = render_step(step)
                elif event.observation:
                    step = event.observation
                    msg, msg_type, icon = render_step(step)
                elif event.agent_event and event.agent_event.final_tape:
                    st.session_state.tape = event.agent_event.final_tape
                if msg:
                    if last_status is not None:
                        last_status.update(state="complete")
                    last_status = None
                    if msg_type == "progress":
                        last_status = st.status(msg)
                    elif msg_type == "code":
                        with st.chat_message("assistant", avatar=":material/code:"):
                            st.code(msg, language="python")
                    elif msg_type == "html":
                        if icon:
                            with st.chat_message("assistant", avatar=f":material/{icon}:"):
                                st.html(msg)
                        else:
                            with st.chat_message("assistant"):
                                st.html(msg)
                    elif msg_type == "markdown":
                        with st.chat_message("assistant"):
                            st.markdown(msg)
                    else:  # default case for "write"
                        with st.chat_message("assistant"):
                            st.write(msg)
                    st.session_state.messages.append({"role": "assistant", "content": msg, "type": msg_type})

        except Exception as e:
            error_msg = f"Failed to solve task: {e}\nStack trace:\n{traceback.format_exc()}"
            print(error_msg)
            st.session_state.tape.metadata.error = str(e)
            with st.chat_message("assistant"):
                st.write(error_msg)


def render_step(step: Step) -> str:
    msg = ""
    msg_type = "html"
    icon = ""
    if step.kind == "plan_thought":
        steps_html = "\n".join(
            [
                f'<div class="step" style="margin: 0; padding: 0 10px; display: flex;">'
                f'<div style="color: #666; margin-right: 10px;">{i + 1}.</div>'
                f"<div>{step}</div>"
                f"</div>"
                for i, step in enumerate(step.plan)
            ]
        )
        msg = f"""
        <div style="font-family: Arial, sans-serif; background: #f8f9fa; border-radius: 8px; padding: 15px; margin: 10px 0;">
            <div style="font-weight: bold; color: #1a73e8; margin-bottom: 10px;">📋 Plan:</div>
            <div style="margin-left: 10px;">
                {steps_html}
            </div>
        </div>
        """
    elif step.kind == "page_observation":
        msg = "Reading web page..."
        msg_type = "progress"
    elif step.kind == "image":
        if step.error:
            msg = f"<div style='color: red;'>{step.error}</div>"
        else:
            msg = "Looking at the image..."
            msg_type = "progress"
    elif step.kind == "computer_observation":
        if not step.base64_image:
            msg = f"""
            <div style="font-family: 'Monaco', 'Menlo', monospace; font-size: 10pt; margin: 10px 0; border-radius: 8px; overflow: hidden;">
                <div style="padding: 15px; background: #f8f9fa; border: 1px solid #eee;">
                    <div style="white-space: pre-wrap;"><code>{step.output}</code></div>
                    {f'<div style="white-space: pre-wrap; color: #dc3545;"><span style="color: #b71c1c;">Error:</span><br><code>{step.error}</code></div>' if step.error else ''}
                </div>
            </div>
            """
            icon = "output"
        else:
            msg = "Looking at the screen..."
            msg_type = "progress"
    elif step.kind == "facts_survey_thought":
        sections = []
        if step.given_facts:
            facts_html = "\n".join([f'<li style="margin: 4px 0;">{fact}</li>' for fact in step.given_facts])
            sections.append(f"""
                <div class="section">
                    <div style="font-weight: bold; color: #1a73e8;">📌 Given Facts:</div>
                    <ul style="margin: 8px 0; padding-left: 20px;">{facts_html}</ul>
                </div>
            """)
        if step.facts_to_lookup:
            facts_html = "\n".join([f'<li style="margin: 4px 0;">{fact}</li>' for fact in step.facts_to_lookup])
            sections.append(f"""
                <div class="section">
                    <div style="font-weight: bold; color: #1a73e8;">🔎 Facts to Look Up:</div>
                    <ul style="margin: 8px 0; padding-left: 20px;">{facts_html}</ul>
                </div>
            """)
        if step.facts_to_derive:
            facts_html = "\n".join([f'<li style="margin: 4px 0;">{fact}</li>' for fact in step.facts_to_derive])
            sections.append(f"""
                <div class="section">
                    <div style="font-weight: bold; color: #1a73e8;">🔄 Facts to Derive:</div>
                    <ul style="margin: 8px 0; padding-left: 20px;">{facts_html}</ul>
                </div>
            """)
        if step.facts_to_guess:
            facts_html = "\n".join([f'<li style="margin: 4px 0;">{fact}</li>' for fact in step.facts_to_guess])
            sections.append(f"""
                <div class="section">
                    <div style="font-weight: bold; color: #1a73e8;">🎲 Facts to Guess:</div>
                    <ul style="margin: 8px 0; padding-left: 20px;">{facts_html}</ul>
                </div>
            """)

        msg = f"""
        <div style="font-family: Arial, sans-serif; background: #f8f9fa; border-radius: 8px; padding: 15px; margin: 10px 0;">
            {"".join(sections)}
        </div>
        """
    elif step.kind == "reasoning_thought":
        msg = step.reasoning
        msg_type = "markdown"
    elif step.kind == "reading_result_thought":
        msg = f'{step.fact_description}\nSupporting quote: "{step.quote_with_fact}"'
    elif step.kind == "assistant_answer":
        if step.success:
            msg = f"""
            <div style="font-family: Arial, sans-serif; margin: 10px 0;">
                <div style="background: #e7f3ff; border-left: 4px solid #1a73e8; padding: 15px; border-radius: 4px;">
                    <div style="font-weight: bold; color: #1a73e8; font-size: 1.1em; margin-bottom: 10px;">
                        🎯 Answer
                    </div>
                    <div style="color: #202124; line-height: 1.5;">
                        {step.answer}
                    </div>
                </div>
            </div>
            """
        else:
            msg = f"""
            <div style="font-family: Arial, sans-serif; margin: 10px 0;">
                <div style="background: #ffeaea; border-left: 4px solid #dc3545; padding: 15px; border-radius: 4px;">
                    <div style="font-weight: bold; color: #dc3545; font-size: 1.1em; margin-bottom: 10px;">
                        ❌ Could Not Find Answer
                    </div>
                    <div style="color: #202124; line-height: 1.5;">
                        {step.answer}
                    </div>
                </div>
            </div>
            """
    elif step.kind == "search_action":
        source_icon = "🌐" if step.source == "web" else "📚" if step.source == "wiki" else "▶️"
        source_name = "Web" if step.source == "web" else "Wikipedia" if step.source == "wiki" else "YouTube"
        msg = f'<div style="padding: 10px; border-left: 3px solid #ccc; background: #f9f9f9;">{source_icon} Searching <b>{source_name}</b> for:<br/><i>"{step.query}"</i></div>'
    elif step.kind == "python_code_action":
        msg = step.code
        msg_type = "code"
    elif step and step.kind == "page_up_action":
        msg = "Scrolling up..."
        msg_type = "progress"
    elif step.kind == "page_down_action":
        msg = "Scrolling down..."
        msg_type = "progress"
    elif step.kind == "mouse_click_at_action":
        msg = f"Clicking {step.element_description}..."
        msg_type = "progress"
    elif step.kind == "run_terminal_command":
        msg = f"$> {step.command}"
        msg_type = "progress"
    elif step.kind == "mouse_hover_action":
        msg = f"Hovering over {step.element_description}..."
        msg_type = "progress"
    elif step.kind in ["input_text_action", "type_text_action", "key_press_action"]:
        msg = f"Typing '{step.text}'..."
        msg_type = "progress"
    elif step.kind in [
        "go_forward_action",
        "go_back_action",
        "click_action",
        "select_option_action",
    ]:
        msg = "Interacting with the computer..."
        msg_type = "progress"
    elif step.kind == "open_url_action":
        msg = f"Opening {step.url}"
        msg_type = "progress"

    elif step.kind == "search_results_observation":
        if step.error:
            msg = step.error
        else:
            results_html = []
            for r in step.serp:
                results_html.append(f"""
                    <div style="margin: 5px 0; padding: 8px; border: 1px solid #eee; border-radius: 4px; background: white;">
                        <div style="line-height: 1.2;">
                            <a href="{r['url']}" style="color: #1a0dab; text-decoration: none; font-size: 0.95em;">{r['title']}</a>
                            <span style="color: #006621; font-size: 0.8em; margin-left: 8px;">
                                {r['url'][:70]}{'...' if len(r['url']) > 70 else ''}
                            </span>
                        </div>
                        <div style="color: #545454; font-size: 0.85em; margin-top: 2px;">
                            {r['content']}
                        </div>
                    </div>
                """)
            msg = '<div style="font-family: Arial, sans-serif;">' + "\n".join(results_html) + "</div>"
        icon = "output"
    elif step.kind == "code_execution_result":
        result = step.result
        status_color = "#28a745" if result.exit_code == 0 else "#dc3545"  # green for success, red for error
        output = result.output.strip() if result.exit_code == 0 else ""
        error = result.output.strip() if result.exit_code != 0 else ""
        msg = f"""
        <div style="font-family: 'Monaco', 'Menlo', monospace; font-size: 10pt; margin: 10px 0; border-radius: 8px; overflow: hidden;">
            <div style="padding: 8px 15px; background: {status_color}; color: white; font-size: 0.9em;">
                Exit Code: {result.exit_code}
            </div>
            <div style="padding: 15px; background: #f8f9fa; border: 1px solid #eee;">
                <div style="white-space: pre-wrap; margin-bottom: {' 15px' if error else '0'};"><code>{output}</code></div>
                {f'<div style="white-space: pre-wrap; color: #dc3545;"><span style="color: #b71c1c;">Error:</span><br><code>{error}</code></div>' if error else ''}
            </div>
        </div>
        """
        icon = "output"
    elif step.kind == "extracted_facts_thought":
        msg = step.extracted_facts
        msg_type = "write"
    else:
        msg = step.llm_dict()
        msg_type = "write"
    return msg, msg_type, icon


def load_from_storage(filename: str) -> str | None:
    """Load data from a file in the storage directory."""
    try:
        file_path = CONFIG_DIR / filename
        if file_path.exists():
            data = file_path.read_text().strip()
            if data:
                return data
    except Exception as e:
        st.write(f"Debug: Error loading {filename}: {e}")
    return None


def save_to_storage(filename: str, data: str) -> None:
    """Save data to a file in the storage directory."""
    try:
        CONFIG_DIR.mkdir(parents=True, exist_ok=True)
        file_path = CONFIG_DIR / filename
        file_path.write_text(data)
        # Ensure only user can read/write the file
        file_path.chmod(0o600)
        st.warning(f"{filename} saved to {CONFIG_DIR}")
    except Exception as e:
        st.write(f"Debug: Error saving {filename}: {e}")


if __name__ == "__main__":
    with initialize(version_base=None, config_path="../../conf", job_name="web_chat"):
        cfg = compose(config_name="web_agent.yaml")
    asyncio.run(main(cfg))



import json
import logging
import os
import re
import time
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime, timezone
from typing import Literal

import requests
from pydantic import BaseModel, ConfigDict, Field
from termcolor import colored

from tapeagents.core import Action, Observation, Prompt
from tapeagents.llms import LLM
from tapeagents.steps import ExplainableAction
from tapeagents.tools.base import Tool
from tapeagents.tools.browser import Fetcher
from tapeagents.tools.simple_browser import SimpleTextBrowser
from tapeagents.tools.tool_cache import cached_tool
from tapeagents.utils import FatalError, response_format

logger = logging.getLogger(__name__)


def web_search_tool(query: str, max_results: int = 5, retry_pause: int = 5, attempts: int = 3) -> list[dict]:
    """
    Search the web for a given query, return a list of search result dictionaries.
    """
    return _web_search(query, max_results=max_results, retry_pause=retry_pause, attempts=attempts)


@cached_tool
def _web_search(query: str, max_results: int = 5, retry_pause: int = 5, attempts: int = 3) -> list[dict]:
    try:
        results = web_search(query, max_results=max_results, retry_pause=retry_pause, attempts=attempts)
    except Exception as e:
        logger.exception(f"Failed to search the web: {e}")
        raise e
    return results


def web_search(
    query: str, time_interval: str = "", max_results: int = 5, retry_pause: int = 2, attempts: int = 3
) -> list[dict]:
    results = []
    while not results and attempts > 0:
        attempts -= 1
        results = serper_search(query, max_results=max_results, time_interval=time_interval)
        if not results:
            logger.warning(f"Empty search results, retrying in {retry_pause} seconds")
            time.sleep(retry_pause)
    if not results:
        raise Exception("Failed to get search results, try to use browser to access the search engine instead")
    return results


def serper_search(query: str, max_results: int = 5, time_interval: str = "") -> list[dict]:
    api_key = os.environ.get("SERPER_API_KEY")
    if not api_key:
        raise FatalError("SERPER_API_KEY env var is not set")
    topic = "videos" if "site:youtube.com" in query else "search"
    payload = {"q": query, "location": "United States", "num": max_results}
    if time_interval:
        payload["tbs"] = f"qdr:{time_interval}"
    headers = {"X-API-KEY": api_key, "Content-Type": "application/json"}
    try:
        response = requests.request(
            "POST", f"https://google.serper.dev/{topic}", headers=headers, data=json.dumps(payload)
        )
        response_dict = response.json()
    except Exception as e:
        logger.exception(f"Serper API error: {e}")
        raise FatalError(f"Failed to get search results: {e}")
    organic = response_dict.get("organic", [])
    videos = response_dict.get("videos", [])
    news = response_dict.get("news", [])
    results = organic + videos + news
    logger.info(f"Search response for query '{query}': code {response.status_code}, {len(results)} results")
    if response.status_code != 200 and len(results) == 0:
        logger.error(f"Search response: {response_dict}")
    return [{"title": r["title"], "url": r["link"], "snippet": r.get("snippet", "")} for r in results[:max_results]]


class SearchAction(Action):
    """
    Action that provides parameters for a search function call.
    Could search in the web, wikipedia or youtube.
    Search results will be ordered by relevance from top to bottom.
    """

    kind: Literal["search_action"] = "search_action"
    source: str = Field(description="source to search in, could be web, wiki or youtube")
    query: str = Field(description="search query")
    time_interval: str = Field(description="time interval for the search, could be h, d, w, m or empty", default="")


class SafeSearchAction(SearchAction):
    kind: Literal["safe_search_action"] = "safe_search_action"
    private_context: list[str] = Field(default_factory=list, exclude=True)


class SearchResultsObservation(Observation):
    kind: Literal["search_results_observation"] = "search_results_observation"
    query: str
    serp: list[dict[str, str]]
    error: str | None = None

    def short_view(self):
        view = self.llm_dict()
        for i, result in enumerate(view["serp"]):
            if "content" in result:
                view["serp"][i]["content"] = result["content"][:100] + "..."
        short = json.dumps(view, indent=2, ensure_ascii=False)
        logger.info(f"SearchResultsObservation long view was {len(self.llm_view())} chars, short view is {len(short)}")
        return short


class SafeSearchResultsObservation(SearchResultsObservation):
    kind: Literal["safe_search_results_observation"] = "safe_search_results_observation"
    safe_search: bool = False
    safe_query: str = ""
    risk_level: int | None = None
    query_rewritten: bool = False
    details: str = ""


class WebSearch(Tool):
    """
    Performs a search in the web, wikipedia or youtube
    """

    action: type[Action] = SearchAction
    observation: type[Observation] = SearchResultsObservation
    cached: bool = True

    def execute_action(self, action: SearchAction) -> SearchResultsObservation:
        if action.source == "wiki":
            query = f"site:wikipedia.org {action.query}"
        elif action.source == "youtube":
            query = f"site:youtube.com {action.query}"
        else:
            query = action.query
        error = None
        results = []
        try:
            results = web_search(query)
        except Exception as e:
            logger.exception(f"Failed to search the web: {e}")
            error = str(e)
        return SearchResultsObservation(query=action.query, serp=results, error=error)


DEFAULT_SAFE_SEARCH_PROMPT = """You are the “Confidential-Data Guardian”.

Confidentiality rules (memorize, never disclose):
• CONFIDENTIAL → unreleased features, codenames/launch dates, internal roadmaps or strategy, private financials, proprietary tech, customer/partner lists, employee PII, security or legal/internal M&A material, anything labeled “Confidential.”
• NOT CONFIDENTIAL → information already public (press releases, launched products, disclosed financials, open-source code, published patents, public testimonials, job-site head-count, industry-standard best practices).

Tasks
1. From the private context, extract ≤ 5 bullet-point “protected_facts” that match the CONFIDENTIAL list.
2. Decide if answering the query verbatim would reveal any protected_fact → set needs_rewrite true/false.
3. If true, create a safe “result_query” that preserves intent but omits protected facts; otherwise return the original query.

Return **only** this JSON (no extra text):

{{
  "protected_facts": ["fact 1", "fact 2", …],   // may be []
  "needs_rewrite": true|false,
  "result_query": "…rewritten or original query…"
}}
"""


class SafeWebSearch(WebSearch):
    observation: type[Observation] = SafeSearchResultsObservation
    llm: LLM
    prompt: str = DEFAULT_SAFE_SEARCH_PROMPT
    max_private_context_len: int = 300000  # characters

    def execute_action(self, action: SafeSearchAction) -> SafeSearchResultsObservation:
        assert isinstance(action.private_context, list)
        query = action.query
        try:
            # Render the prompt
            private_context_str = "\n".join(action.private_context)

            if len(private_context_str) > self.max_private_context_len:
                logger.warning(
                    f"Private context is too long ({len(private_context_str)} chars), truncating to {self.max_private_context_len} chars"
                )
                private_context_str = private_context_str[: self.max_private_context_len]

            content = f"CONFIDENTIAL DATA:\n{private_context_str}\n\nORIGINAL SEARCH QUERY: {query}"
            prompt = Prompt(messages=[{"role": "system", "content": self.prompt}, {"role": "user", "content": content}])
            privacy_mitigator_output = self.llm.generate(prompt).get_text()

            # Get privacy level
            level = None
            match = re.search(r"PRIVACY RISK:\s*(.+)", privacy_mitigator_output)
            if match:
                level_str = match.group(1).strip()[:1]
                if level_str in ["1", "2", "3", "4", "5"]:
                    level = int(level_str)

            # Get query
            new_query = None
            match = re.search(r"NEW SEARCH QUERY:\s*(.+)", privacy_mitigator_output)
            if match:
                new_query = match.group(1).strip()
                if new_query.lower() == "none":
                    new_query = None

            # If both are present and level >= 3, keep the new query
            if level and new_query and level >= 3:
                query_rewritten = True
                logger.warning(
                    f'\nSAFE_SEARCH: Risk Level {level} Rewriting old query to new query: "{query}" -> "{new_query}"'
                )
            else:
                # If no new query is provided, keep the original query
                query_rewritten = False
                new_query = query
                logger.warning(f'\nSAFE_SEARCH: Risk Level {level} Keeping original query "{query}"')

            results = web_search(new_query, time_interval=action.time_interval)
            result_obs = SafeSearchResultsObservation(
                safe_search=True,
                safe_query=new_query,
                query_rewritten=query_rewritten,
                query=query,
                risk_level=level,
                details=privacy_mitigator_output,
                serp=results,
            )
        except Exception as e:
            logger.exception(f"Failed to search the web: {e}")
            result_obs = SafeSearchResultsObservation(query=action.query, safe_search=True, error=str(e), serp=[])

        return result_obs


class SuperSearch(WebSearch):
    """
    Performs a search in the web or wikipedia.
    Retrieves the whole content of each page.
    """

    enriched_results: int = 3
    page_viewport_size: int = 32000

    def model_post_init(self, __context):
        self._browser = SimpleTextBrowser(viewport_size=self.page_viewport_size)
        return super().model_post_init(__context)

    def execute_action(self, action: SearchAction) -> SearchResultsObservation:
        obs = super().execute_action(action)
        for i, result in enumerate(obs.serp[: self.enriched_results]):
            text, total_pages, error = self._browser.get_page(result["url"])
            logger.info(f"Fetched {len(text)} chars from the search result {i}")
            if not error:
                obs.serp[i]["content"] = text
            else:
                logger.warning(f"Failed to fetch page {result['url']}: {error}")
        return obs


class SearchTask(BaseModel):
    model_config = ConfigDict(extra="forbid")
    section: str
    facts_to_discover: str
    queries: list[str]


class SearchExtractAction(
    ExplainableAction,
    SearchTask,
):
    """
    Action that provides parameters for a search function call.
    Could search in the web, wikipedia or youtube.
    Search results will be ordered by relevance from top to bottom.
    """

    kind: Literal["search_extract_action"] = "search_extract_action"  # type: ignore


class SearchAndExtract(Action):
    kind: Literal["search_and_extract"] = "search_and_extract"
    main_task: str
    instructions: str
    tasks: list[SearchTask]
    time_interval: str
    skip_urls: list[str]
    private_context: list[str] = Field(default_factory=list, exclude=True)  # hide so it's not dumped


class PageFacts(BaseModel):
    model_config = ConfigDict(extra="forbid")
    publication_date: str = Field(description="Date in YYYY-MM-DD format, empty string if not available")
    facts: list[str]


class WebPageData(BaseModel):
    model_config = ConfigDict(extra="forbid")
    url: str
    title: str
    content: str | list[str]
    prompt_id: str
    fetch_date_time: str
    date: str | None = None


class ExtractedFactsObservation(Observation):
    kind: Literal["extracted_facts_observation"] = "extracted_facts_observation"
    page_facts: dict[str, list[WebPageData]]

    def llm_view(self, indent=2):
        task_facts = []
        for task, pages in self.page_facts.items():
            prefix = f"Topic: {task}:\n<FACTS>"
            page_strs: list[str] = []
            for page in pages:
                page_facts = (
                    "\n".join([f"- {f}" for f in page.content]) if isinstance(page.content, list) else page.content
                )
                page_strs.append(
                    f"Page [{page.title}][{page.url}]:\nPublication date: {page.date or 'Unknown'}. Extraction date: {page.fetch_date_time}. Facts:\n{page_facts}\n----"
                )
            task_facts.append(prefix + "\n\n".join(page_strs) + "\n</FACTS>")
        facts = "\n\n".join(task_facts)
        return f"<FACTS_COLLECTION>Extracted facts:\n{facts}\n</FACTS_COLLECTION>"

    def short_view(self):
        tasks = "\n".join(self.page_facts.keys())
        return f"Extracted facts for tasks:\n{tasks}"


class SearchResult(BaseModel):
    model_config = ConfigDict(extra="forbid")
    task_id: int
    query_id: int
    n: int
    title: str
    url: str
    snippet: str
    text: str = ""

    # Additional fields for safe search
    query: str = ""
    safe_query: str = ""
    safe_search: bool = False
    query_rewritten: bool = False
    risk_level: int | None = None
    details: str = ""


class MultiSearchExtract(Tool):
    """
    Performs a search in the web and retrieves the whole content of each page. Then extracts the relevant information.
    """

    llm: LLM
    action: type[Action] = SearchAndExtract
    observation: type[Observation] = ExtractedFactsObservation
    cached: bool = True
    top_k: int = 3
    max_workers: int = 20
    search_timeout: int = 10
    fetch_timeout: int = 30
    extract_timeout: int = 60
    extract_prefix: str = "Your should extract all relevant information from the page.\n\nTASK: "
    safe_search: bool = False
    safe_search_prompt: str = DEFAULT_SAFE_SEARCH_PROMPT

    def model_post_init(self, __context):
        if self.safe_search:
            self._search_tool = SafeWebSearch(llm=self.llm, cached=self.cached, prompt=self.safe_search_prompt)
        else:
            self._search_tool = WebSearch(cached=self.cached)
        return super().model_post_init(__context)

    def execute_action(self, action: SearchAndExtract) -> ExtractedFactsObservation:
        search_results = self.search(action)
        fetch_results = self.fetch(search_results)
        extracted_facts = self.extract(action, fetch_results)
        logger.info(f"Extracted facts from {sum([len(p) for p in extracted_facts.values()])} pages.")
        return ExtractedFactsObservation(page_facts=extracted_facts)

    def search(self, action: SearchAndExtract) -> list[SearchResult]:
        def search_query(i: int, j: int, query: str) -> list[SearchResult]:
            results = []
            if self.safe_search:
                results_obs: SafeSearchResultsObservation = self._search_tool.run(
                    SafeSearchAction(
                        source="web",
                        query=query,
                        private_context=action.private_context,
                        time_interval=action.time_interval,
                    )
                )  # type: ignore
                serp = results_obs.serp[: self.top_k]
                for n, r in enumerate(serp):
                    results.append(
                        SearchResult(
                            task_id=i,
                            query_id=j,
                            query=query,
                            safe_query=results_obs.safe_query,
                            safe_search=True,
                            n=n,
                            title=r["title"],
                            url=r["url"],
                            snippet=r["snippet"],
                        )
                    )
            else:
                serp = self._search_tool.run(
                    SearchAction(source="web", query=query, time_interval=action.time_interval)
                ).serp[: self.top_k]  # type: ignore
                for n, r in enumerate(serp):
                    if r["url"] in action.skip_urls:
                        logger.info(f"Skipping URL {r['url']}")
                        continue
                    results.append(
                        SearchResult(
                            task_id=i,
                            query_id=j,
                            query=query,
                            n=n,
                            title=r["title"],
                            url=r["url"],
                            snippet=r["snippet"],
                        )
                    )
            return results

        search_tasks = [
            (task_id, query_id, query)
            for task_id, search_task in enumerate(action.tasks)
            for query_id, query in enumerate(search_task.queries)
        ]
        executor = ThreadPoolExecutor(max_workers=self.max_workers)
        futures = [executor.submit(search_query, *s) for s in search_tasks]
        results: list[SearchResult] = []
        try:
            for future in as_completed(futures, timeout=self.search_timeout):
                results += future.result()
        except Exception as e:
            logger.error(f"Error occurred while processing web search future: {e}")
        executor.shutdown(wait=False)
        logger.info(f"Got {len(results)} search results for {len(search_tasks)} queries.")
        return results

    def fetch(self, search_results: list[SearchResult]) -> list[SearchResult]:
        fetcher = Fetcher()
        texts = {}
        urls = list(set([result.url for result in search_results]))
        executor = ThreadPoolExecutor(max_workers=self.max_workers)
        futures = [executor.submit(fetcher.fetch_for_llm, url) for url in urls]
        try:
            for future in as_completed(futures, timeout=self.fetch_timeout):
                url, text = future.result()
                texts[url] = text
                logger.info(f"Fetched {len(texts)} out of {len(urls)} pages")
        except Exception as e:
            logger.error(f"Error occurred while processing fetch future: {e}")
        executor.shutdown(wait=False)
        logger.info(f"Fetched {len(texts)} pages")
        for i in range(len(search_results)):
            search_results[i].text = texts.get(search_results[i].url, "")
        return search_results

    def extract(self, action: SearchAndExtract, fetch_results: list[SearchResult]) -> dict[str, list[WebPageData]]:
        extract_tasks = []
        for fr in fetch_results:
            task = action.tasks[fr.task_id].section
            facts_to_discover = action.tasks[fr.task_id].facts_to_discover
            query = action.tasks[fr.task_id].queries[fr.query_id]
            prefix = f"{self.extract_prefix}{task} (part of higher-level task {action.main_task})\nFacts to discover: {facts_to_discover}\nSearch query that led to the page: {query}\n\nPage content:\n\n"
            postfix = f"\n\nData extraction instructions:\n{action.instructions}\nIf you found the creation date of the document, put it in the first line. If the page is empty or contains only message about blocking the access, return only one word ERROR and nothing else."
            msg = f"{prefix}{fr.text}{postfix}"
            logger.info(f"Page: {fr.url}, prompt length {len(msg)} chars")
            prompt = Prompt(messages=[{"role": "user", "content": msg}])
            extract_tasks.append((task, fr.url, fr.title, prompt))

        def extract_page_data(task: str, url: str, title: str, prompt: Prompt) -> tuple[str, WebPageData]:
            page_data_content = self.llm.generate(prompt).get_text()
            if page_data_content.startswith("ERROR"):
                logger.warning(f"Page {url} empty or blocked")
                return "", None

            msg = f"Convert the following text into a structured JSON object:\n\n{page_data_content}\n\nIf there is no date in the first line of the text, leave date field empty. Do not split the tables into multiple facts, put every table into single fact string."
            page_data_json = self.llm.generate(
                Prompt(messages=[{"role": "user", "content": msg}], response_format=response_format(PageFacts))
            ).get_text()
            try:
                page_facts = PageFacts.model_validate_json(page_data_json)
            except Exception as e:
                logger.exception(f"Failed to produce structured list of facts: {e}")
                return None, None  # type: ignore
            logger.info(
                colored(
                    f"Completed extraction for page: {url}\nDate {page_facts.publication_date}, facts: {page_facts.facts}",
                    "green",
                )
            )
            return task, WebPageData(
                url=url,
                title=title,
                content=page_facts.facts,
                prompt_id=prompt.id,
                date=page_facts.publication_date,
                fetch_date_time=datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S"),
            )

        data_per_task = defaultdict(list)
        executor = ThreadPoolExecutor(max_workers=self.max_workers)
        futures = [executor.submit(extract_page_data, *et) for et in extract_tasks]
        extracted = 0
        try:
            for future in as_completed(futures, timeout=self.extract_timeout):
                task, page_data = future.result()
                extracted += 1
                logger.info(f"Extracted {extracted} out of {len(extract_tasks)} pages")
                if not task:
                    continue
                data_per_task[task].append(page_data)
        except Exception as e:
            logger.error(f"Error occurred while processing extract future: {e}")
        executor.shutdown(wait=False)
        return data_per_task


EXTRACT_INSTRUCTIONS = """
Create markdown document with all the facts and claims from the provided document.
First try to find the creation date of the document in its content. If it is not available, skip it and not mention.
Then collect all facts, claims events and definitions mentioned in the document and produce a complete list.
Things that are relevant to the current task should come first.
If there are any tables, use markdown to represent them.
Fact is any numerical or categorical information about a person, place, event, product, object or a company.
Claim is any statement made by a person that is cited in the web page. Claims should be attributed to the person who made them and represented as verbatim quotes.
When providing quotes, ensure that the quote includes enough context to be understood.
Do not append the word "Fact" or "Claim" to the beginning of sentences.
Important! Respond with the plain text, do not include any JSON or code.
Do not output anything besides what I asked in this message.
"""


class SearchExtract(MultiSearchExtract):
    """
    Performs a search in the web for multiple queries and retrieves the whole content of each page.
    Then extracts the relevant information.
    """

    action: type[Action] = SearchExtractAction
    observation: type[Observation] = ExtractedFactsObservation

    def execute_action(self, action: SearchExtractAction) -> ExtractedFactsObservation:
        multi_action = SearchAndExtract(
            main_task="perform web search and extract facts",
            instructions=EXTRACT_INSTRUCTIONS,
            tasks=[action],
            time_interval="",
            skip_urls=[],
            private_context=[],
        )
        search_results = self.search(multi_action)
        fetch_results = self.fetch(search_results)
        extracted_facts = self.extract(multi_action, fetch_results)
        logger.info(f"Extracted facts from {sum([len(p) for p in extracted_facts.values()])} pages.")
        return ExtractedFactsObservation(page_facts=extracted_facts)



import os
import sys
from itertools import zip_longest

import gradio as gr

from tapeagents.core import LLMCall
from tapeagents.io import load_tapes
from tapeagents.renderers.basic import BasicRenderer
from tapeagents.utils import diff_strings

from ..eval import calculate_accuracy, tape_correct
from ..steps import GaiaTape


class TapeDiffGUI:
    """
    GUI for comparing two set of tapes from different runs.
    Assuming that the tapes are stored in JSON files with the same structure and produced from the same tasks.
    So the tape N from the first file is compared with the tape N from the second file.
    """

    def __init__(
        self,
        renderer: BasicRenderer,
        fname: str,
        fname2: str,
        file_extension: str = ".json",
    ):
        self.renderer = renderer
        self.dirname = os.path.dirname(fname)
        self.dirname2 = os.path.dirname(fname2)
        self.fname = os.path.basename(fname)
        self.fname2 = os.path.basename(fname2)
        self.file_extension = file_extension

    def load_tapes(self, fname: str) -> tuple[list[GaiaTape], dict[str, LLMCall], str]:
        tapes: list[GaiaTape] = load_tapes(GaiaTape, fname)  # type: ignore
        llm_calls = {}  # TODO: load prompts
        acc, n_solved = calculate_accuracy(tapes)
        header = f"Accuracy: {acc:.2f}%, {n_solved} out of {len(tapes)}"
        return tapes, llm_calls, header

    def tape_name(self, n: int, tape: GaiaTape, tape2: GaiaTape | None) -> str:
        correct = tape_correct(tape)
        task = tape.metadata.task
        mark = "+" if correct else "-"
        if tape2:
            mark += "+" if tape_correct(tape2) else "-"
        if task["file_name"]:
            mark += "[f]"
        return f"{mark}L{task['Level']}:{n+1}: {task['Question'][:32]}"

    def on_load(self):
        files = sorted([f for f in os.listdir(self.dirname) if f.endswith(self.file_extension)])
        files2 = sorted([f for f in os.listdir(self.dirname2) if f.endswith(self.file_extension)])
        file_selector = gr.Dropdown(files, label="File A", value=self.fname)  # type: ignore
        file_selector2 = gr.Dropdown(files2, label="File B", value=self.fname2)  # type: ignore
        return file_selector, file_selector2

    def update(self, fname: str, fname2: str, n: int, first_time: bool = False):
        tapes, llm_calls, header = self.load_tapes(os.path.join(self.dirname, fname))
        tapes2, llm_calls2, header2 = self.load_tapes(os.path.join(self.dirname2, fname2))
        tape_names = []

        # prepare list of tape names
        for i, tape in enumerate(tapes):
            name = self.tape_name(i, tape, tapes2[i] if i < len(tapes2) else None)
            tape_names.append((name, i))
        tape_names2 = []
        for i, tape in enumerate(tapes2):
            name = self.tape_name(i, tape, tapes[i] if i < len(tapes) else None)
            tape_names2.append((name, i))
        if not first_time:
            tape_names = gr.Dropdown(tape_names, label="Tape A", value=n)
            tape_names2 = gr.Dropdown(tape_names2, label="Tape B", value=n)

        # render pair of tapes
        style = "<style>.basic-renderer-box { width: 600px !important; }</style>"
        html = f"""{self.renderer.style}{style}<table class="diff" cellspacing="0" cellpadding="0"><tr class="diff">
            <td class="diff"><h2 style="padding-left: 2em;">{header}</h2></td>
            <td class="diff"><h2 style="padding-left: 2em;">{header2}</h2></td></tr>"""
        for i, (step, step2) in enumerate(zip_longest(tapes[n].steps, tapes2[n].steps)):
            prompt_text = ""
            diff_class = "diff"
            if step is None:
                step_str = ""
            else:
                step_str = self.renderer.render_step(step, i, folded=False)
                prompt_id = step.metadata.prompt_id
                if prompt_id and prompt_id in llm_calls:
                    prompt_text = self.renderer.render_llm_call(llm_calls[prompt_id])
                    step_str = prompt_text + step_str
            if step2 is None:
                step2_str = ""
            else:
                step2_str = self.renderer.render_step(step2, i, folded=False)
                prompt_id2 = step2.metadata.prompt_id
                if step and step != step2:
                    diff_class = "diff_highlight"
                    if step.kind == step2.kind:
                        # highlight differences in step B
                        step2_str = diff_strings(
                            self.renderer.render_step(step, i, folded=False), step2_str, use_html=True, by_words=True
                        )
                if prompt_id2 in llm_calls2:
                    prompt_text2 = self.renderer.render_llm_call(llm_calls2[prompt_id2])
                    if prompt_text and prompt_text != prompt_text2:
                        # highlight differences in prompt B
                        prompt_text2 = diff_strings(prompt_text, prompt_text2, use_html=True, by_words=True)
                    step2_str = prompt_text2 + step2_str
            html += f'<tr class="diff"><td class="diff">{step_str}</td><td class="{diff_class}">{step2_str}</td></tr>'
        html += "</table>"

        return html, tape_names, tape_names2

    def launch(self, server_name: str = "0.0.0.0"):
        tape = 0
        with gr.Blocks() as blocks:
            html, tape_names, tape_names2 = self.update(self.fname, self.fname2, tape, first_time=True)
            with gr.Row():
                with gr.Column(scale=4):
                    file_selector = gr.Dropdown([], label="File A")
                    selector = gr.Dropdown(tape_names, label="Tape A", value=tape)  # type: ignore
                with gr.Column(scale=4):
                    file_selector2 = gr.Dropdown([], label="File B")
                    selector2 = gr.Dropdown(tape_names2, label="Tape B", value=tape)  # type: ignore
            with gr.Row():
                tape_view = gr.HTML(html)

            selector.change(
                fn=self.update,
                inputs=[file_selector, file_selector2, selector],
                outputs=[tape_view, selector, selector2],
            )
            selector2.change(
                fn=self.update,
                inputs=[file_selector, file_selector2, selector2],
                outputs=[tape_view, selector, selector2],
            )
            file_selector.change(
                fn=self.update,
                inputs=[file_selector, file_selector2, selector],
                outputs=[tape_view, selector, selector2],
            )
            file_selector2.change(
                fn=self.update,
                inputs=[file_selector, file_selector2, selector],
                outputs=[tape_view, selector, selector2],
            )
            blocks.load(self.on_load, inputs=None, outputs=[file_selector, file_selector2])
        blocks.launch(server_name=server_name)


if __name__ == "__main__":
    assert len(sys.argv) == 3, "Usage: python -m scripts.tape_diff <fname1> <fname2>"
    gui = TapeDiffGUI(BasicRenderer(), sys.argv[1], sys.argv[2])
    gui.launch()



from tapeagents.core import LLMCall, LLMOutput

from .base import LLM, LLMEvent, LLMStream
from .claude import Claude
from .lite import LiteLLM
from .mock import MockLLM
from .replay import ReplayLLM
from .trainable import TrainableLLM

__all__ = [
    "LLM",
    "LLMCall",
    "LLMOutput",
    "LLMStream",
    "LLMEvent",
    "LiteLLM",
    "ReplayLLM",
    "TrainableLLM",
    "MockLLM",
    "Claude",
]



import logging
import os
import sys
from pathlib import Path

from tapeagents.core import Tape
from tapeagents.io import load_tapes
from tapeagents.renderers.camera_ready_renderer import CameraReadyRenderer
from tapeagents.tape_browser import TapeBrowser

from ..tape import FormFillerTape
from ..user_simulator_agent import UserSimulatorTape

# Syntax: python -m examples.form_filler.scripts.tape_browser <path_to_tapes_folder> [<path_to_sqlite_db>] [<port>]

tape_dir = Path(sys.argv[1])

if len(sys.argv) >= 3:
    os.environ["TAPEAGENTS_SQLITE_DB"] = sys.argv[2]

logger = logging.getLogger(__name__)


# Configure to show info messages in the console
logging.basicConfig(level=logging.INFO)


class RecursiveTapeBrowser(TapeBrowser):
    def load_tapes(self, fname: str) -> list[Tape]:
        tapes = []
        try:
            tapes = load_tapes(FormFillerTape, fname)
            logger.info(f"{len(tapes)} FormFillerTape tapes loaded from {fname}")
        except Exception as e:
            try:
                tapes = load_tapes(UserSimulatorTape, fname)
                logger.info(f"{len(tapes)} UserSimulatorTape loaded from {fname}")
            except Exception as e2:
                logger.error(f"Could not load tapes from {fname}.")
                logger.error(f"Tried loading as FormFillerTape, got this error: {e}")
                logger.error(f"Tried loading as UserSimulatorTape, got this error: {e2}")
        return tapes

    def get_tape_files(self) -> list[str]:
        files = sorted(
            [
                os.path.join(root, f)
                for root, _, filenames in os.walk(self.tapes_folder)
                for f in filenames
                if f.endswith(self.file_extension) and ".hydra" not in root
            ]
        )
        assert files, f"No files found in {self.tapes_folder}"
        logger.info(f"{len(files)} files found in {self.tapes_folder}")
        indexed = 0
        nonempty_files = []
        for i, file in enumerate(files):
            if file in self.files:
                continue  # already indexed
            tapes = []
            try:
                tapes = self.load_tapes(file)
            except Exception as e:
                logger.error(f"Could not load tapes from {file}: {e}")
                continue
            if not len(tapes):
                logger.warning(f"File {file} does not contain any known tapes, skip")
                continue
            for j, tape in enumerate(tapes):
                tape_id = tape.metadata.id
                parent_id = tape.metadata.parent_id
                if tape_id:
                    if tape_id in self.tape_index and self.tape_index[tape_id] != (i, j):
                        raise ValueError(
                            f"Duplicate tape id {tape_id}. Both in {self.tape_index[tape_id]} and {(i, j)}"
                        )
                    indexed += 1
                    self.tape_index[tape_id] = (i, j)
                    if parent_id:
                        if parent_id not in self.tape_children:
                            self.tape_children[parent_id] = []
                        self.tape_children[parent_id].append(tape_id)
            nonempty_files.append(file)
        logger.info(f"Indexed {indexed} new tapes, index size: {len(self.tape_index)}")
        logger.info(f"{len(self.tape_children)} tapes with children found")
        return nonempty_files


browser = RecursiveTapeBrowser(Tape, sys.argv[1], CameraReadyRenderer(), file_extension=".yaml")
browser.launch(port=7680 if len(sys.argv) < 4 else int(sys.argv[3]))



from itertools import takewhile
from typing import Generator

from examples.rl_webagent.utils import StepRepeatMonitor
from tapeagents.agent import Agent
from tapeagents.core import (
    Action,
    AgentStep,
    LLMOutputParsingFailureAction,
    PartialStep,
    SetNextNode,
    Step,
    StopStep,
)
from tapeagents.dialog_tape import UserStep
from tapeagents.llms import LLM, LLMStream
from tapeagents.nodes import StandardNode
from tapeagents.tools.simple_browser import PageObservation
from tapeagents.utils import FatalError

from .prompts import PromptRegistry
from .steps import (
    ReasoningThought,
    WebAgentAction,
    WebAgentStep,
    WebTape,
    WebTask,
)


class WebNode(StandardNode):
    max_retries: int = (
        3  # max number of times to retry the current node if the last step is an LLMOutputParsingFailureAction
    )
    current_retries: int = 0  # current number of retries for the current node
    max_same_action: int = 4  # max number of times to repeat the same action
    max_chars_page_observation: int = 2000  # max number of characters to keep in PageObservation["text"]

    def tape_to_messages(self, tape: WebTape, steps_description: str) -> list[dict]:
        """
        Converts a Tape object and steps description into a list of messages for LLM conversation.

        Modifications from the original StandardNode:
        - If the last n steps are LLMOutputParsingFailureAction, put the guidance before the error steps
          and add a new guidance message that asks to retry.
        - Do not render older LLMOutputParsingFailureAction steps that got solved.
        - Always render WebTask observations as they contain the instructions
        - Truncate PageObservation steps up to `max_chars_page_observation` chars (instead of 100 by default).

        Args:
            tape (Tape): A Tape object containing conversation steps.
            steps_description (str): A description of the conversation steps.

        Returns:
            list[dict]: A list of dictionaries representing the conversation messages.
                       Each dictionary contains 'role' and 'content' keys.
                       Roles can be 'system', 'user', or 'assistant'.
                       The system prompt is always the first message.
                       If steps_description is provided, it's added as a user message.
                       Messages from tape are added with roles based on step type.
                       If guidance exists, it's added as the final user message.
        """
        messages: list[dict] = []
        if self.system_prompt:
            messages.append({"role": "system", "content": self.system_prompt})
        if steps_description:
            messages.append({"role": "user", "content": steps_description})
        # ignore the last n consecutive parsing error steps for now, will add them later
        last_parsing_error_steps: list[LLMOutputParsingFailureAction] = []
        for step in reversed(tape.steps):
            if isinstance(step, LLMOutputParsingFailureAction):
                last_parsing_error_steps.append(step)
            else:
                break
        # put back the last n parsing error steps in the original order
        last_parsing_error_steps.reverse()
        # remove all the parsing error steps
        cleaned_steps = [step for step in tape.steps if not isinstance(step, LLMOutputParsingFailureAction)]
        # get the positions of all observations to trim the old ones
        page_observation_idx = [i for i, step in enumerate(cleaned_steps) if isinstance(step, PageObservation)]
        for i, step in enumerate(cleaned_steps):
            role = "assistant" if isinstance(step, AgentStep) else "user"
            if isinstance(step, PageObservation):
                if i not in page_observation_idx[-self.trim_obs_except_last_n :]:
                    # skip old page observations
                    continue
                else:
                    view = step.short_view(max_chars=self.max_chars_page_observation)
                    # view = f"Page Observation: ```json\n{view}\n```"
            elif isinstance(step, WebTask):
                view = f"Task: {step.task}"
            elif isinstance(step, UserStep):
                view = step.content
            elif isinstance(step, ReasoningThought):
                view = step.reasoning
            else:
                view = step.llm_view()
            messages.append({"role": role, "content": view})
        # add the initial guidance message
        if self.guidance:
            messages.append({"role": "user", "content": self.guidance})
        # add the last n parsing error steps and the new guidance message that asks to retry
        for step in last_parsing_error_steps:
            messages.append({"role": "assistant", "content": step.llm_view()})
            messages.append(
                {
                    "role": "user",
                    "content": "You made a mistake. Look at your generation (in 'llm_output') and the error message (in 'error') and please try again.",
                }
            )
        return messages

    def generate_steps(
        self, agent: Agent, tape: WebTape, llm_stream: LLMStream
    ) -> Generator[Step | PartialStep, None, None]:
        """
        Generates a sequence of steps based on the LLM stream output.

        This method processes the output from a language model stream and converts it into a series of steps.
        It handles the parsing of completions and post-processing of steps.

        Modifications from the original StandardNode:
        - If the last step is an LLMOutputParsingFailureAction, retry the current node
          **up to max_retries times**.
        - If the last action is repeated more than **max_same_action times**, stop the task.

        Args:
            agent (Any): The agent instance that will execute the steps.
            tape (Tape): The tape object containing the execution context and history.
            llm_stream (LLMStream): The stream of language model outputs to process.

        Yields:
            Union[Step, PartialStep]: Individual steps generated from the LLM stream output.

        Raises:
            FatalError: If no completions are generated from the LLM stream.

        Note:
            - If the node has a next_node defined and the final step is not a StopStep,
              it will yield a SetNextNode step to continue the execution flow.
        """
        # get the last action and the number of repetitions
        previous_actions = [step for step in tape.steps if isinstance(step, Action)]
        last_action = previous_actions[-1] if previous_actions else None
        n_last_actions = len(list(takewhile(lambda x: x == last_action, reversed(previous_actions))))
        action_monitor = StepRepeatMonitor(
            last_step=last_action, repeat_count=n_last_actions, max_repeats=self.max_same_action
        )

        new_steps = []
        for event in llm_stream:
            if not event.output:
                continue
            if event.output.content:
                new_steps += list(self.parse_completion(event.output.content))
            if event.output.tool_calls and self.use_function_calls:
                new_steps += [self.tool_call_to_step(agent, tool_call) for tool_call in event.output.tool_calls]

            for i, step in enumerate(new_steps):
                step = self.postprocess_step(tape, new_steps[:i], step)
                yield step
                # check that the new step is not a repetition of the last action up to max_same_action times
                if isinstance(step, Action) and action_monitor.should_stop(step):
                    raise FatalError(f"Max same action reached! {step}")
                # if the last step was an LLMOutputParsingFailureAction, retry the current node up to max_retries times
                if isinstance(step, LLMOutputParsingFailureAction):
                    if self.current_retries < self.max_retries:
                        retry_step = SetNextNode(next_node=self.name)
                        yield retry_step
                        self.current_retries += 1
                        new_steps.append(retry_step)
                        break
                    else:
                        raise FatalError(f"Max retries reached for node {self.name}, parsing error: {step.llm_view()}!")
                else:
                    self.current_retries = 0
        if not new_steps:
            raise FatalError("No completions!")
        if (
            self.next_node
            and not isinstance(new_steps[-1], StopStep)
            and not any(isinstance(step, SetNextNode) for step in new_steps)
        ):
            yield SetNextNode(next_node=self.next_node)


class WebAgent(Agent):
    @classmethod
    def create(cls, llm: LLM, max_iterations: int = 4):
        return super().create(
            llm,
            nodes=[
                WebNode(
                    name="set_goal",
                    guidance=PromptRegistry.start,
                    system_prompt=PromptRegistry.system_prompt,
                    steps_prompt=PromptRegistry.allowed_steps,
                    steps=ReasoningThought,
                    trim_obs_except_last_n=3,  # keep the last 3 observations from the tape in prompt messages
                    max_chars_page_observation=3000,  # keep up to 3000 chars in PageObservation steps
                ),
                WebNode(
                    name="reflect",
                    guidance=PromptRegistry.reflect,
                    system_prompt=PromptRegistry.system_prompt,
                    steps_prompt=PromptRegistry.allowed_steps,
                    steps=WebAgentStep,
                    trim_obs_except_last_n=3,  # keep the last 3 observations from the tape in prompt messages
                    max_chars_page_observation=3000,  # keep up to 3000 chars in PageObservation steps
                ),
                WebNode(
                    name="act",
                    guidance=PromptRegistry.act,
                    system_prompt=PromptRegistry.system_prompt,
                    steps_prompt=PromptRegistry.allowed_steps,
                    steps=WebAgentAction,
                    trim_obs_except_last_n=3,  # keep the last 3 observations from the tape in prompt messages
                    max_chars_page_observation=3000,  # keep up to 3000 chars in PageObservation steps
                    next_node="reflect",
                ),
            ],
            max_iterations=max_iterations,
            store_llm_calls=True,
        )



from typing import Callable

import torch
import transformers
from torch.utils.data.dataloader import DataLoader

from .context import get_accelerator, logger
from .types import TrainingMetrics


def evaluate(
    args,
    model: transformers.PreTrainedModel,
    eval_dataloader: DataLoader,
):
    model.eval()
    losses = []
    with torch.no_grad():
        for step, batch in enumerate(eval_dataloader):
            outputs = model(**batch)
            loss = outputs.loss.repeat(args.valid_batch_size)
            losses.append(get_accelerator().gather(loss).detach())  # type: ignore
            if args.max_eval_steps > 0 and step >= args.max_eval_steps:
                break
    model.train()
    return torch.mean(torch.cat(losses)).item()


def evaluate_and_get_metrics(args, model, eval_dataloader, dev_dataloader, training_metrics: TrainingMetrics):
    if eval_dataloader:
        logger.info("Evaluating model")
        training_metrics.eval_loss = evaluate(args, model, eval_dataloader)
        if dev_dataloader:
            training_metrics.dev_loss = evaluate(args, model, dev_dataloader)
        else:
            training_metrics.dev_loss = 0
        if training_metrics.eval_loss < training_metrics.best_eval_loss:
            training_metrics.best_eval_loss = training_metrics.eval_loss
            training_metrics.best_completed_steps = training_metrics.completed_steps
    return training_metrics


def dummy_eval_callback(config_name: str) -> Callable:
    def dummy(*args, **kwargs):
        return {}

    return dummy



import tempfile
from pathlib import Path

from tapeagents.batch import batch_main_loop
from tapeagents.dialog_tape import DialogTape, SystemStep, UserStep
from tapeagents.environment import EmptyEnvironment
from tapeagents.io import stream_yaml_tapes
from tapeagents.llms import LLM, TrainableLLM

from .llama_agent import LLAMAChatBot


def try_batch_main_loop(llm: LLM):
    agent = LLAMAChatBot.create(llms=llm)
    tape = DialogTape(
        context=None,
        steps=[
            SystemStep(
                content="Respond to the user using the style of Shakespeare books. Be very brief, 50 words max."
            ),
            UserStep(content="Hello, how are you?"),
        ],
    )

    path = Path(tempfile.mktemp())
    with stream_yaml_tapes(path) as dumper:
        for new_tape in batch_main_loop(agent, [tape] * 10, EmptyEnvironment()):
            dumper.save(new_tape)

    with open(path) as src:
        print(src.read())


if __name__ == "__main__":
    try_batch_main_loop(
        TrainableLLM(
            base_url="https://api.together.xyz",
            model_name="meta-llama/Llama-3-8b-chat-hf",
            parameters=dict(temperature=0.7, max_tokens=512),
        )
    )



from tapeagents.agent import Agent, Node
from tapeagents.core import Prompt
from tapeagents.dialog_tape import AssistantStep, DialogTape, UserStep
from tapeagents.llms import LiteLLM, LLMStream
from tapeagents.prompting import tape_to_messages

llm = LiteLLM(model_name="gpt-4o-mini")


class MainNode(Node):
    def make_prompt(self, agent: Agent, tape: DialogTape) -> Prompt:
        # Render the whole tape into the prompt, each step is converted to message
        return Prompt(messages=tape_to_messages(tape))

    def generate_steps(self, agent: Agent, tape: DialogTape, llm_stream: LLMStream):
        # Generate single tape step from the LLM output messages stream.
        yield AssistantStep(content=llm_stream.get_text())


agent = Agent[DialogTape].create(llm, nodes=[MainNode()])
start_tape = DialogTape(steps=[UserStep(content="Tell me about Montreal in 3 sentences")])
final_tape = agent.run(start_tape).get_final_tape()  # agent will start executing the first node
print(f"Final tape: {final_tape.model_dump_json(indent=2)}")



import logging
import os
from typing import Literal, Union

from pydantic import Field

from tapeagents.agent import Agent
from tapeagents.core import (
    Action,
    FinalStep,
    LLMOutputParsingFailureAction,
    Observation,
    SetNextNode,
    Tape,
    Thought,
)
from tapeagents.environment import Environment
from tapeagents.io import save_json_tape
from tapeagents.llms import LLM
from tapeagents.nodes import StandardNode
from tapeagents.orchestrator import main_loop
from tapeagents.tools.calculator import calculate

logger = logging.getLogger(__name__)


#### Steps ####
class Task(Observation):
    kind: Literal["task"] = "task"
    task: str


class ReasoningThought(Thought):
    """
    Thoughts produced by the agent during the reasoning process.
    """

    kind: Literal["reasoning_thought"] = "reasoning_thought"
    reasoning: str = Field(description="chain of thoughts")


class UseCalculatorAction(Action):
    """
    Action to use calculator to find the new value. This math expression should be a single line. You can use exp, cos, sin, tan, abs, trunc, sgn, round
    """

    kind: Literal["use_calculator_action"] = "use_calculator_action"
    expression: str = Field(description="math expression to calculate")


class CalculationResultObservation(Observation):
    kind: Literal["calculation_result_observation"] = "calculation_result_observation"
    result: str


class ActionExecutionFailure(Observation):
    kind: Literal["action_execution_failure"] = "action_execution_failure"
    error: str


class AnswerAction(FinalStep):
    """
    Action that provides the final answer to the user after completing the task. Should be produced when the agent has finished the task.
    """

    kind: Literal["final_answer_action"] = "final_answer_action"
    text: str = Field(description="final answer to the user")
    value: int | float | None = Field(description="numerical value of the answer or null if solution is not found")


MathAgentStep = (UseCalculatorAction, ReasoningThought, AnswerAction)


MathTape = Tape[
    None,
    Union[
        Task,
        ReasoningThought,
        UseCalculatorAction,
        CalculationResultObservation,
        ActionExecutionFailure,
        LLMOutputParsingFailureAction,
        AnswerAction,
        SetNextNode,
    ],
]

#### Prompts ####

SYSTEM_PROMPT = """You are the genius math agent. Help user solve math problems.
Your role is to understand user queries and respond in a helpful and accurate manner.
Keep your replies concise and direct. Prioritize clarity and avoid over-elaboration.
"""

ALLOWED_STEPS = """
You are allowed to produce ONLY steps with the following json schemas:
{allowed_steps}
Do not reproduce schema when producing the steps, use it as a reference.
"""

HINTS = """
Important considerations:
- Always produce only one step at a time.
- Step kind is always lowercase and underscore separated.
- DO NOT OUTPUT ANYTHING BESIDES THE JSON. It will break the system that processes the output.
"""

START_TASK_GUIDANCE = f"Let's think step by step using reasoning and calculations.\n{HINTS}"


#### Agent and Environment ####
class MathAgent(Agent):
    @classmethod
    def create(cls, llm: LLM):
        return super().create(
            llm,
            nodes=[
                StandardNode(
                    name="start",
                    system_prompt=SYSTEM_PROMPT,
                    steps_prompt=ALLOWED_STEPS,
                    steps=MathAgentStep,
                    guidance=START_TASK_GUIDANCE,
                ),
                StandardNode(
                    name="default",
                    system_prompt=SYSTEM_PROMPT,
                    steps_prompt=ALLOWED_STEPS,
                    steps=MathAgentStep,
                    guidance=HINTS,
                    next_node="default",
                ),
            ],
            max_iterations=2,
        )


class MathEnvironment(Environment):
    def __init__(self) -> None:
        super().__init__()

    def react(self, tape: MathTape) -> MathTape:
        actions = [step for step in tape.steps[-tape.metadata.n_added_steps :] if isinstance(step, Action)]
        for action in actions:
            if isinstance(action, LLMOutputParsingFailureAction):
                continue
            try:
                match action:
                    case UseCalculatorAction():
                        observation = CalculationResultObservation(result=calculate(action.expression, {}))
                        tape = tape.append(observation)
                    case _:
                        raise Exception(f"Unknown action: {type(action)}")
            except Exception as e:
                tape = tape.append(ActionExecutionFailure(error=str(e)))
                break
        return tape


def solve_task(agent: Agent, env: Environment, task: dict, tape_file: str = "") -> Tape:
    tmp_tape_file = f"{tape_file}.tmp" if tape_file else None
    start_step = Task(task=task["question"])
    tape = MathTape(steps=[start_step], context=None)
    metadata = task.copy()

    for event in main_loop(agent, tape, env, max_loops=30):
        step = None
        if event.agent_event and event.agent_event.step:
            step = event.agent_event.step
        elif event.observation:
            step = event.observation
        if step:
            tape = tape.append(step)  # type: ignore
            if tmp_tape_file:
                save_json_tape(tape, tmp_tape_file)
    if tmp_tape_file:
        os.unlink(tmp_tape_file)  # remove temporary file
    metadata["solved"] = False
    if isinstance(tape[-1], AnswerAction) and task["value"] == tape[-1].value:  # type: ignore
        metadata["solved"] = True
    tape.metadata.result = metadata
    return tape


def extract_result_value(sample) -> dict:
    sample = dict(sample)
    expected_result = str(sample["answer"]).rsplit("####", maxsplit=1)[-1].strip().replace(",", "")
    if expected_result.isdigit():
        expected_result = int(expected_result)
    else:
        expected_result = float(expected_result)
    sample["value"] = expected_result
    return sample



import json
import logging
from collections import Counter
from datetime import datetime
from pathlib import Path
from typing import Generator

import hydra
from hydra.utils import instantiate
from omegaconf import DictConfig
from tqdm import tqdm

from tapeagents.agent import Agent
from tapeagents.core import Error, TapeMetadata
from tapeagents.io import load_tapes, stream_yaml_tapes
from tapeagents.orchestrator import main_loop
from tapeagents.parallel_processing import lazy_thread_pool_processor

from ..environment import FormFillerEnvironment
from ..error import UnknownError
from ..student import StudentAgent
from ..tape import (
    FormFillerAgentMetadata,
    FormFillerContext,
    FormFillerTape,
    FormFillerUserMetadata,
)

logger = logging.getLogger(__name__)


def run_formfiller_agent(
    tape: FormFillerTape, agent: Agent[FormFillerTape]
) -> tuple[FormFillerTape, FormFillerTape | Exception]:
    """
    Run the agent on the given tape
    Returns the input tape and (the predicted tape or an exception)
    """
    try:
        # assert this tape is a user tape
        assert isinstance(
            tape.metadata, FormFillerUserMetadata
        ), f"Expected FormFillerUserMetadata as input tape, got {type(tape.metadata)}. Tape: {tape.metadata.id}"
        # create environment
        assert isinstance(
            tape.context, FormFillerContext
        ), f"Expected FormFillerContext, got {type(tape.context)}. Tape: {tape.metadata.id}"
        env = FormFillerEnvironment.from_spec(tape.context.env_spec)
        # run main_loop
        predicted_tape = main_loop(agent, tape, env).get_final_tape()
        # update metadata accordingly
        assert isinstance(predicted_tape.metadata, TapeMetadata)
        predicted_tape.metadata = FormFillerAgentMetadata(
            **predicted_tape.metadata.model_dump(),
            last_user_tape_id=tape.metadata.id,
        )
        # update context
        assert isinstance(predicted_tape.context, FormFillerContext)
        predicted_tape.context.date = datetime.now().strftime("%Y-%m-%d")
        return tape, predicted_tape
    except Exception as e:
        return tape, e


@hydra.main(version_base=None, config_path="../conf", config_name="run_formfiller_agent")
def main(cfg: DictConfig):
    agent = instantiate(cfg.agent)

    agent_type = isinstance(agent, StudentAgent) and "student" or "teacher"
    logger.info(f"Agent type: {agent_type}")

    user_dialogues_path = cfg.user_dialogues_path
    tapes = load_tapes(FormFillerTape, user_dialogues_path)

    def tape_gen() -> Generator[FormFillerTape, None, None]:
        for tape in tqdm(tapes):
            yield tape  # type: ignore

    predicted_tapes: list[FormFillerTape] = []
    failure_tapes: list[FormFillerTape] = []
    for result in lazy_thread_pool_processor(
        stream=tape_gen(),
        worker_func=lambda tape: run_formfiller_agent(tape, agent),
        n_workers=cfg.n_workers,
    ):
        if isinstance(result, Exception):
            logger.exception(f"Error while running lazy_thread_pool_processor: {result}", exc_info=result)
            continue
        else:
            input_tape, result = result
            if isinstance(result, Exception):
                logger.exception(f"Error while running run_formfiller_agent: {result}", exc_info=result)
                predicted_tape = input_tape.model_copy(
                    update=dict(steps=input_tape.steps + [UnknownError(message=str(result))])
                )
            else:
                predicted_tape = result

        if predicted_tape.last_action is None or any(isinstance(step, Error) for step in predicted_tape.steps):
            failure_tapes.append(predicted_tape)
        else:
            predicted_tapes.append(predicted_tape)

    output_path = Path(cfg.output_path)  # must be a directory
    output_path.mkdir(exist_ok=True, parents=True)

    # save predicted tapes
    stats = Counter()
    with stream_yaml_tapes(output_path / f"{agent_type}_predicted_tapes.yaml") as saver:
        for tape in tqdm(predicted_tapes, desc="Saving predicted tapes"):
            stats[tape.last_action.kind] += 1
            saver.save(tape)
    print(stats)

    # save failure tapes
    error_stats = Counter()
    if failure_tapes:
        with stream_yaml_tapes(output_path / f"{agent_type}_failure_tapes.yaml") as saver:
            for tape in tqdm(failure_tapes, desc="Saving failure tapes"):
                error_stats[tape.last_action.kind] += 1
                saver.save(tape)
        print(error_stats)

    # save stats: counters of last step kinds
    with open(output_path / f"{agent_type}_stats.json", "w") as f:
        json.dump(
            {
                "predicted_tapes_stats": stats,
                "failure_tapes_stats": error_stats,
            },
            f,
            indent=2,
        )


if __name__ == "__main__":
    main()



import logging
import os
import sys

import numpy as np
from datasets import load_dataset
from termcolor import colored
from tqdm import tqdm

from tapeagents.io import save_json_tape
from tapeagents.llms import TrainableLLM

from .math_agent import (
    MathAgent,
    MathEnvironment,
    extract_result_value,
    solve_task,
)

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def main(exp_path: str, attempts: int = 1):
    dataset = load_dataset("openai/gsm8k", "main", split="train")
    samples = [s for s in dataset]
    np.random.seed(42)
    np.random.shuffle(samples)  # type: ignore
    logging.info(f"Loaded {len(samples)} samples")

    llm = TrainableLLM(
        base_url="https://api.together.xyz",
        model_name="meta-llama/Meta-Llama-3.1-70B-Instruct-Turbo",
        tokenizer_name="meta-llama/Meta-Llama-3.1-70B-Instruct",
        parameters=dict(temperature=0.2),
    )
    agent = MathAgent.create(llm)
    env = MathEnvironment()

    tapes_dir = os.path.join(exp_path, "tapes")
    logger.info(f"Saving tapes to {tapes_dir}")
    os.makedirs(tapes_dir, exist_ok=True)
    os.environ["TAPEAGENTS_SQLITE_DB"] = os.path.join(exp_path, "tapedata.sqlite")

    solved = []
    for i, sample in enumerate(tqdm(samples)):
        sample = extract_result_value(sample)
        for j in range(attempts):
            tape_file = os.path.join(tapes_dir, f"task{i}_attempt{j+1}.json")
            if os.path.exists(tape_file):
                logger.info(f"Task {i} attempt {j+1} already solved, skipping")
                continue
            try:
                tape = solve_task(agent, env, sample, tape_file)
                solved.append(int(tape.metadata.result["solved"]))
                save_json_tape(tape, tape_file)
            except Exception as e:
                logger.error(colored(f"Failed to solve task, attempt {j+1}: {e}", "red"))
                solved.append(0)
        if i % 10 == 0 and i > 0:
            logger.info(f"{i}: Current accuracy: {np.mean(solved):.3f}, prompt tokens used: {agent.llm.token_count}")
    logger.info(f"Accuracy: {np.mean(solved):.3f}, prompt tokens used: {agent.llm.token_count}")
    logger.info(f"{len(solved)} tapes saved to {tapes_dir}")


if __name__ == "__main__":
    exp_path = sys.argv[1] if len(sys.argv) > 1 else "gsm8k/tuning/llama31_70b_train"
    main(exp_path)



import logging
from typing import Callable, Generator, Mapping

import gradio as gr
import uvicorn
import yaml
from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles

from tapeagents.agent import Agent
from tapeagents.core import Tape
from tapeagents.environment import Environment
from tapeagents.observe import (
    get_latest_tape_id,
    observe_tape,
    retrieve_tape,
    retrieve_tape_llm_calls,
)
from tapeagents.orchestrator import MainLoopEvent, main_loop
from tapeagents.renderers import render_agent_tree
from tapeagents.renderers.basic import BasicRenderer

logger = logging.getLogger(__name__)


class Studio:
    def __init__(
        self,
        agent: Agent,
        tape: Tape,
        renderers: BasicRenderer | Mapping[str, BasicRenderer],
        environment: Environment | None = None,
        transforms: Mapping[str, Callable[[Tape], Tape]] | None = None,
    ):
        self.renderers = (
            {renderers.__class__.__name__: renderers} if isinstance(renderers, BasicRenderer) else renderers
        )
        self.environment = environment
        self.transforms = transforms or {}
        self.original_agent = agent
        self.original_tape = tape

        with gr.Blocks(title="TapeAgent Studio") as blocks:
            tape_state = gr.State(tape.model_dump())
            agent_state = gr.State(agent.model_dump())
            with gr.Row():
                with gr.Column(scale=1):
                    tape_data = gr.Textbox(
                        "", max_lines=15, label="Raw Tape content", info="Press Enter to rerender the tape"
                    )
                    pop = gr.Number(1, label="Pop N last steps", info="Press Enter to proceed")
                    keep = gr.Number(0, label="Keep N first steps", info="Press Enter to proceed")
                    load = gr.Textbox(max_lines=1, label="Load tape by id", info="Press Enter to load tape")
                    choices = list(self.renderers.keys())
                    renderer_choice = gr.Dropdown(choices=choices, value=choices[0], label="Choose the tape renderer")  # type: ignore
                    if transforms:
                        transform_choice = gr.Dropdown(list(transforms.keys()), label="Run a transform")
                with gr.Column(scale=3):
                    tape_render = gr.HTML("")
                with gr.Column(scale=1):
                    gr.TextArea(render_agent_tree(agent), label="Agents Hierarchy", max_lines=20)
                    agent_config = gr.Textbox(
                        "",
                        max_lines=15,
                        label="Agents configuration",
                        info="Press Enter to update the agent",
                    )
                    run_agent = gr.Button("Run Agent")
                    if environment:
                        run_enviroment = gr.Button("Run Environment")
                        run_loop = gr.Button("Run Loop")
                    else:
                        run_enviroment = None
                        run_loop = None

            render_tape = (self.render_tape, [renderer_choice, tape_state], [tape_data, tape_render])

            blocks.load(*render_tape).then(self.render_agent, [agent_state], [agent_config])

            # Tape controls
            tape_data.submit(
                lambda data: tape.model_validate(yaml.safe_load(data)).with_new_id(),
                [tape_data],
                [tape_state],
            ).then(*render_tape).then(lambda tape_dict: observe_tape(self.validate_tape(tape_dict)), [tape_state], [])
            pop.submit(self.pop_n_steps, [tape_state, pop], [tape_state]).then(*render_tape)
            keep.submit(self.keep_n_steps, [tape_state, keep], [tape_state]).then(*render_tape)
            load.submit(self.load_tape, [tape_state, load], [tape_state]).then(*render_tape)
            renderer_choice.change(*render_tape)
            if transforms:
                transform_choice.change(
                    lambda transform, tape: self.transforms[transform](tape),
                    [transform_choice, tape_state],
                    [tape_state],
                ).then(*render_tape)

            # Agent controls
            agent_config.submit(self.update_agent, [agent_config], [agent_state]).then(lambda: gr.Info("Agent updated"))
            run_agent.click(
                self.run_agent, [renderer_choice, agent_state, tape_state], [tape_state, tape_data, tape_render]
            )
            if environment:
                assert run_enviroment and run_loop
                run_enviroment.click(self.run_environment, [tape_state], [tape_state]).then(*render_tape)
                run_loop.click(
                    self.run_main_loop, [renderer_choice, agent_state, tape_state], [tape_state, tape_data, tape_render]
                )

        self.blocks = blocks

    def validate_tape(self, tape_dict: dict) -> Tape:
        return self.original_tape.model_validate(tape_dict)

    def pop_n_steps(self, tape_dict: dict, n: int) -> Tape:
        tape = self.validate_tape(tape_dict)
        if n > len(tape):
            raise gr.Error(f"Cannot pop {n} steps from tape with {len(tape)} steps")
        return tape[:-n]

    def keep_n_steps(self, tape_dict: dict, n: int) -> Tape:
        tape = self.validate_tape(tape_dict)
        if n > len(tape):
            raise gr.Error(f"Cannot keep {n} steps from tape with {len(tape)} steps")
        return tape[:n]

    def transform_tape(self, transform: str, tape_dict: dict) -> Tape:
        tape = self.validate_tape(tape_dict)
        result = self.transforms[transform](tape)
        observe_tape(result)
        return result

    def load_tape(self, tape_dict: dict, tape_id: str) -> Tape:
        tape = self.validate_tape(tape_dict)
        if not tape_id:
            tape_id = get_latest_tape_id()
        result = retrieve_tape(type(tape), tape_id)
        if not result:
            raise gr.Error(f"No tape found with id {tape_id}")
        return result

    def render_tape(self, renderer_name: str, tape_dict: dict) -> tuple[str, str]:
        tape = self.validate_tape(tape_dict)
        renderer = self.renderers[renderer_name]
        llm_calls = retrieve_tape_llm_calls(tape)
        return (
            yaml.safe_dump(tape.model_dump(), sort_keys=False),
            renderer.style + renderer.render_tape(tape, llm_calls),
        )

    def validate_agent(self, agent_dict: dict) -> Agent:
        return self.original_agent.update(agent_dict)

    def render_agent(self, agent_dict: dict) -> str:
        agent = self.validate_agent(agent_dict)
        try:
            agent_config = yaml.dump(agent.model_dump(), sort_keys=False)
        except Exception as e:
            logger.exception(f"Failed to get agent configuration: {e}")
            agent_config = "Failed to get agent configuration"
        return agent_config

    def update_agent(self, config: str) -> dict:
        return self.original_agent.update(yaml.safe_load(config)).model_dump()

    def run_agent(
        self, renderer_name: str, agent_dict: dict, tape_dict: dict
    ) -> Generator[tuple[Tape, str, str], None, None]:
        agent = self.validate_agent(agent_dict)
        start_tape = self.validate_tape(tape_dict)
        for event in agent.run(start_tape):
            if tape := event.partial_tape or event.final_tape:
                observe_tape(tape)
                yield (tape,) + self.render_tape(renderer_name, tape)

    def run_environment(self, tape: Tape) -> Tape:
        assert self.environment
        start_tape = self.validate_tape(tape)
        return self.environment.react(start_tape)

    def run_main_loop(
        self, renderer_name: str, agent_dict: dict, tape_dict: dict
    ) -> Generator[tuple[Tape, str, str], None, None]:
        assert self.environment
        agent = self.validate_agent(agent_dict)
        start_tape = self.validate_tape(tape_dict)
        last_tape = start_tape
        for event in main_loop(agent, start_tape, self.environment):
            if ae := event.agent_event:
                if ae.step:
                    logger.info(f"added step {type(ae.step).__name__}")
                if tape := ae.partial_tape or ae.final_tape:
                    observe_tape(tape)
                    last_tape = tape
                    yield (tape,) + self.render_tape(renderer_name, tape)
            elif event.observation:
                last_tape = last_tape.append(event.observation)
                yield (last_tape,) + self.render_tape(renderer_name, last_tape)
            elif isinstance(event, MainLoopEvent):
                pass
            else:
                raise ValueError("Unexpected event", event)

    def launch(
        self, server_name: str = "0.0.0.0", port=7860, debug: bool = False, static_dir: str = "", *args, **kwargs
    ):
        if static_dir:
            logger.info(f"Starting FastAPI server with static dir {static_dir}")
            # mount Gradio app to FastAPI app
            app = FastAPI()
            app.mount("/static", StaticFiles(directory=static_dir), name="static")
            app = gr.mount_gradio_app(app, self.blocks, path="/")
            uvicorn.run(app, host=server_name, port=port)
        else:
            self.blocks.launch(server_name=server_name, debug=debug, *args, **kwargs)



import datetime
import logging
import sys

from tapeagents.environment import CodeExecutionEnvironment
from tapeagents.llms import LLM, LiteLLM
from tapeagents.orchestrator import main_loop
from tapeagents.renderers.pretty import PrettyRenderer
from tapeagents.studio import Studio
from tapeagents.team import TeamAgent, TeamTape
from tapeagents.tools.container_executor import ContainerExecutor

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")

# Prompt borrowed from https://github.com/microsoft/autogen, MIT License
DEFAULT_TEAM_AGENT_SYSTEM_MESSAGE = """You are a helpful AI assistant.
Solve tasks using your coding and language skills.
In the following cases, suggest python code (in a python coding block) or shell script (in a sh coding block) for the user to execute.
    1. When you need to collect info, use the code to output the info you need, for example, browse or search the web, download/read a file, print the content of a webpage or a file, get the current date/time, check the operating system. After sufficient info is printed and the task is ready to be solved based on your language skill, you can solve the task by yourself.
    2. When you need to perform some task with code, use the code to perform the task and output the result. Finish the task smartly.
Solve the task step by step if you need to. If a plan is not provided, explain your plan first. Be clear which step uses code, and which step uses your language skill.
When using code, you must indicate the script type in the code block. The user cannot provide any other feedback or perform any other action beyond executing the code you suggest. The user can't modify your code. So do not suggest incomplete code which requires users to modify. Don't use a code block if it's not intended to be executed by the user.
If you want the user to save the code in a file before executing it, put # filename: <filename> inside the code block as the first line. Don't include multiple code blocks in one response. Do not ask users to copy and paste the result. Instead, use 'print' function for the output when relevant. Check the execution result returned by the user.
If the result indicates there is an error, fix the error and output the code again. Suggest the full code instead of partial code or code changes. If the error can't be fixed or if the task is not solved even after the code is executed successfully, analyze the problem, revisit your assumption, collect additional info you need, and think of a different approach to try.
When you find an answer, verify the answer carefully. Include verifiable evidence in your response if possible.
"""


def try_chat(llm: LLM, studio: bool):
    # equilavent of https://microsoft.github.io/autogen/docs/tutorial/introduction
    org = TeamAgent.create_initiator(
        name="UserProxy",
        llm=llm,
        system_prompt="",
        teammate=TeamAgent.create(
            name="Assistant",
            system_prompt=DEFAULT_TEAM_AGENT_SYSTEM_MESSAGE,
            llm=llm,
        ),
        max_calls=3,
        init_message="compute 5 fibonacci numbers",
        execute_code=True,
    )
    start_tape = TeamTape(context=None, steps=[])
    now = f"{datetime.datetime.now():%Y%m%d%H%M%S}"
    env = CodeExecutionEnvironment(ContainerExecutor(work_dir=f"outputs/chat_code/{now}"))
    if studio:
        Studio(org, start_tape, PrettyRenderer(), env).launch()
    else:
        loop = main_loop(org, start_tape, env)
        for event in loop:
            if ae := event.agent_event:
                if ae.step:
                    print(ae.step.model_dump_json(indent=2))
            else:
                assert event.observation
                print(event.observation.model_dump_json(indent=2))


if __name__ == "__main__":
    llm = LiteLLM(model_name="gpt-4o")
    if len(sys.argv) == 2:
        if sys.argv[1] == "studio":
            try_chat(llm, studio=True)
        else:
            raise ValueError()
    elif len(sys.argv) == 1:
        try_chat(llm, studio=False)
    else:
        raise ValueError()



"""
Multiagent building blocks for orchestrating multiple agents in a team.
"""

from __future__ import annotations

import logging
from typing import Generator, Generic

from pydantic import ConfigDict
from typing_extensions import Self

from tapeagents.agent import DEFAULT, Agent, AgentStep, Node
from tapeagents.core import FinalStep, Pass, Prompt, SetNextNode, StepMetadata, Tape, TapeType
from tapeagents.environment import CodeExecutionResult, ExecuteCode
from tapeagents.llms import LLM, LLMStream
from tapeagents.nodes import CallSubagent, RespondIfNotRootNode
from tapeagents.tools.container_executor import extract_code_blocks
from tapeagents.view import Broadcast, Call, Respond, TapeViewStack

logger = logging.getLogger(__name__)


TeamTape = Tape[None, Call | Respond | Broadcast | FinalStep | SetNextNode | ExecuteCode | CodeExecutionResult | Pass]


class ActiveTeamAgentView:
    def __init__(self, agent: TeamAgent, tape: TeamTape):
        """
        ActiveTeamAgentView contains the ephemeral state computed from the tape. This class extracts the data relevant to
        the given agent and also computes some additional information from it, e.g. whether the agent
        should call the LLM to generate a message or respond with an already available one.
        """
        view = TapeViewStack.compute(tape)
        self.messages = view.messages_by_agent[agent.full_name]
        self.last_non_empty_message = next((m for m in reversed(self.messages) if m.content), None)
        self.node = agent.select_node(tape)
        self.steps = view.top.steps
        self.steps_by_kind = view.top.steps_by_kind
        self.exec_result = self.steps[-1] if self.steps and isinstance(self.steps[-1], CodeExecutionResult) else None
        self.should_generate_message = (
            isinstance(self.node, (CallNode, RespondNode))
            and self.messages
            and not self.exec_result
            and "system" in agent.templates
        )
        self.should_stop = (
            agent.max_calls and (agent.max_calls and len(self.steps_by_kind.get("call", [])) >= agent.max_calls)
        ) or (self.messages and ("TERMINATE" in self.messages[-1].content))


class TeamAgent(Agent[TeamTape]):
    """
    Agent designed to work in the team with similar other agents performing different kinds
    """

    max_calls: int | None = None
    init_message: str | None = None

    model_config = ConfigDict(use_enum_values=True)

    @classmethod
    def create(
        cls,
        name: str,
        system_prompt: str | None = None,
        llm: LLM | None = None,
        execute_code: bool = False,
    ):  # type: ignore
        """
        Create a simple agent that can execute code, think and respond to messages
        """
        return cls(
            name=name,
            templates={"system": system_prompt} if system_prompt else {},
            llms={DEFAULT: llm} if llm else {},
            nodes=([ExecuteCodeNode()] if execute_code else []) + [RespondNode()],  # type: ignore
        )

    @classmethod
    def create_team_manager(
        cls,
        name: str,
        subagents: list[Agent[TeamTape]],
        llm: LLM,
        templates: dict[str, str],
        max_calls: int = 1,
    ):
        """
        Create a team manager that broadcasts the last message to all subagents, selects one of them to call, call it and
        responds to the last message if the termination message is not received.
        """
        return cls(
            name=name,
            subagents=subagents,
            nodes=[
                BroadcastLastMessageNode(),
                SelectAndCallNode(),
                RespondOrRepeatNode(next_node="broadcast_last_message"),
            ],
            max_calls=max_calls,
            templates=templates,
            llms={DEFAULT: llm},
        )

    @classmethod
    def create_initiator(
        cls,
        name: str,
        teammate: Agent[TeamTape],
        init_message: str,
        system_prompt: str = "",
        llm: LLM | None = None,
        max_calls: int = 1,
        execute_code: bool = False,
    ):
        """
        Create an agent that sets the team's initial message and calls the team manager
        """
        nodes = []
        if execute_code:
            nodes = [ExecuteCodeNode(), CallNode(), TerminateOrRepeatNode(next_node="execute_code")]
        else:
            nodes = [CallNode(), TerminateOrRepeatNode(next_node="call")]
        return cls(
            name=name,
            templates={
                "system": system_prompt,
            },
            llms={DEFAULT: llm} if llm else {},
            subagents=[teammate],
            nodes=nodes,
            max_calls=max_calls,
            init_message=init_message,
        )


class BroadcastLastMessageNode(Node):
    name: str = "broadcast_last_message"

    def generate_steps(
        self, agent: TeamAgent, tape: TeamTape, llm_stream: LLMStream
    ) -> Generator[AgentStep, None, None]:
        view = ActiveTeamAgentView(agent, tape)
        recipients = agent.get_subagent_names()
        last_step = view.messages[-1]
        from_ = last_step.metadata.agent.split("/")[-1]
        match last_step:
            case Call():
                yield Broadcast(content=last_step.content, from_=from_, to=list(recipients))
            case Respond():
                recipients = [name for name in recipients if name != last_step.metadata.agent.split("/")[-1]]
                yield Broadcast(
                    content=view.messages[-1].content,
                    from_=from_,
                    to=list(recipients),
                )
            case Broadcast(metadata=StepMetadata(node=self.name)):
                pass
            case _:
                assert False


class CallNode(Node):
    name: str = "call"

    def make_prompt(self, agent: TeamAgent, tape: TeamTape) -> Prompt:
        p = Prompt()
        view = ActiveTeamAgentView(agent, tape)
        if view.should_generate_message:
            system = [{"role": "system", "content": agent.templates["system"]}]
            p = Prompt(messages=system + _llm_messages_from_tape(agent, tape))
        return p

    def generate_steps(
        self, agent: TeamAgent, tape: TeamTape, llm_stream: LLMStream
    ) -> Generator[AgentStep, None, None]:
        view = ActiveTeamAgentView(agent, tape)
        # if last node
        (other,) = agent.subagents
        if view.should_generate_message:
            yield Call(agent_name=other.name, content=llm_stream.get_text())
        elif view.exec_result:
            yield Call(agent_name=other.name, content=_exec_result_message(agent, tape))
        else:
            assert agent.init_message and not view.messages
            yield Call(agent_name=other.name, content=agent.init_message)


class SelectAndCallNode(Node):
    name: str = "select_and_call"

    def make_prompt(self, agent: TeamAgent, tape: TeamTape) -> Prompt:
        subagents = ", ".join(agent.get_subagent_names())
        select_before = [
            {
                "role": "system",
                "content": agent.templates["select_before"].format(subagents=subagents),
            }
        ]
        select_after = [
            {
                "role": "system",
                "content": agent.templates["select_after"].format(subagents=subagents),
            }
        ]
        return Prompt(messages=select_before + _llm_messages_from_tape(agent, tape) + select_after)

    def generate_steps(
        self, agent: TeamAgent, tape: TeamTape, llm_stream: LLMStream
    ) -> Generator[AgentStep, None, None]:
        callee_name = llm_stream.get_text()
        # check if the callee is an existing subagent
        _ = agent.find_subagent(callee_name)
        yield Call(agent_name=callee_name)


class ExecuteCodeNode(Node):
    name: str = "execute_code"

    def generate_steps(self, agent: TeamAgent, tape: Tape, llm_stream: LLMStream) -> Generator[AgentStep, None, None]:
        assert not llm_stream
        view = ActiveTeamAgentView(agent, tape)
        if view.last_non_empty_message is None:
            yield Pass()
        elif code := extract_code_blocks(view.last_non_empty_message.content):
            yield ExecuteCode(code=code)
        else:
            yield Pass()


class RespondNode(Node):
    name: str = "respond"

    def make_prompt(self, agent: TeamAgent, tape: TeamTape) -> Prompt:
        p = Prompt()
        view = ActiveTeamAgentView(agent, tape)
        if view.should_generate_message:
            system = [{"role": "system", "content": agent.templates["system"]}]
            p = Prompt(messages=system + _llm_messages_from_tape(agent, tape))
        return p

    def generate_steps(
        self, agent: TeamAgent, tape: TeamTape, llm_stream: LLMStream
    ) -> Generator[AgentStep, None, None]:
        view = ActiveTeamAgentView(agent, tape)
        if view.should_generate_message:
            yield Respond(content=llm_stream.get_text())
        elif view.exec_result:
            yield Respond(content=_exec_result_message(agent, tape))
        else:
            logger.info(
                f"Agent {agent.full_name} had to respond with an empty message."
                f" You might want to optimize your orchestration logic."
            )
            yield Respond()


class TerminateOrRepeatNode(Node):
    name: str = "terminate_or_repeat"
    next_node: str

    def generate_steps(
        self, agent: TeamAgent, tape: TeamTape, llm_stream: LLMStream
    ) -> Generator[AgentStep, None, None]:
        assert not llm_stream
        view = ActiveTeamAgentView(agent, tape)
        if view.should_stop:
            yield FinalStep(reason="Termination message received")
        else:
            yield SetNextNode(next_node=self.next_node)


class RespondOrRepeatNode(Node):
    name: str = "respond_or_repeat"
    next_node: str

    def generate_steps(
        self, agent: TeamAgent, tape: TeamTape, llm_stream: LLMStream
    ) -> Generator[AgentStep, None, None]:
        view = ActiveTeamAgentView(agent, tape)
        if view.should_stop:
            yield Respond()
        else:
            yield SetNextNode(next_node=self.next_node)


def _exec_result_message(agent: TeamAgent, tape: TeamTape) -> str:
    view = ActiveTeamAgentView(agent, tape)
    exec_result_message = ""
    if view.exec_result:
        if output := view.exec_result.result.output.strip():
            exec_result_message = f"I ran the code `{'`, `'.join(view.exec_result.result.code_files)}` and got the following output:\n\n{output}"
        else:
            exec_result_message = f"I ran the code `{'`, `'.join(view.exec_result.result.code_files)}`, the exit code was {view.exec_result.result.exit_code}."
    return exec_result_message


def _llm_messages_from_tape(agent: TeamAgent, tape: TeamTape) -> list[dict[str, str]]:
    view = ActiveTeamAgentView(agent, tape)
    llm_messages = []
    for step in view.messages:
        match step:
            # When we make the LLM messages, we use "kind" == "user" for messages
            # originating from other agents, and "kind" == "assistant" for messages by this agent.
            case Call() if step.metadata.agent == agent.full_name:
                # I called someone
                llm_messages.append({"role": "assistant", "content": step.content})
            case Call():
                # someone called me
                # we exclude empty call messages from the prompt
                if not step.content:
                    continue
                llm_messages.append(
                    {
                        "role": "user",
                        "content": step.content,
                        "name": step.metadata.agent.split("/")[-1],
                    }
                )
            case Respond() if step.metadata.agent == agent.full_name:
                # I responded to someone
                llm_messages.append({"role": "assistant", "content": step.content})
            case Respond():
                # someone responded to me
                who_returned = step.metadata.agent.split("/")[-1]
                llm_messages.append({"role": "user", "content": step.content, "name": who_returned})
            case Broadcast():
                llm_messages.append({"role": "user", "content": step.content, "name": step.from_})
    return llm_messages


class Chain(Agent[TapeType], Generic[TapeType]):
    """Calls agents sequentially. Copies thoughts of previous agents for the next agents."""

    @classmethod
    def create(cls, nodes: list[CallSubagent], **kwargs) -> Self:
        subagents = []
        for node in nodes:
            subagents.append(node.agent)
        return super().create(nodes=nodes + [RespondIfNotRootNode()], subagents=subagents, **kwargs)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Callable, Dict, List, Literal

from pydantic import BaseModel, Field, TypeAdapter

from tapeagents.core import Action, Observation
from tapeagents.tools.base import StatefulTool


class GetMemoryTitlesAction(Action):
    """
    Action to get the titles for all the memories.
    """

    kind: Literal["get_memory_titles_action"] = "get_memory_titles_action"


class GetMemoryAction(Action):
    """
    Action to get a specific memory
    """

    kind: Literal["get_memory_action"] = "get_memory_action"
    key: str = Field(description="A descriptive key for the memory")


class UpsertMemoryAction(Action):
    """
    Action to insert or update a memory
    """

    kind: Literal["upsert_memory_action"] = "upsert_memory_action"
    key: str = Field(description="A descriptive key for the memory")
    title: str = Field(description="title for the memory")
    data: str = Field(description="data for the memory")


class MemoryObservation(Observation):
    kind: Literal["memory_observation"] = "memory_observation"
    key: str = Field(description="A descriptive key for the memory")
    title: str = Field(description="title for the memory")
    data: str = Field(description="data for the memory")


class MemoryTitle(BaseModel):
    key: str = Field(description="A descriptive key for the memory")
    title: str = Field(description="title for the memory")


class MemoryTitlesObservation(Observation):
    kind: Literal["memory_titles_observation"] = "memory_titles_observation"
    titles: List[MemoryTitle] = Field(description="titles and keys for the memories")


class Memory(BaseModel):
    key: str
    title: str
    data: str


class MemoryStorage(ABC):
    @abstractmethod
    def get(self, key: str) -> Memory:
        pass

    @abstractmethod
    def get_all(self) -> List[Memory]:
        pass

    @abstractmethod
    def upsert(self, memory: Memory):
        pass

    @abstractmethod
    def close():
        pass


class InMemoryStorage(MemoryStorage):
    _storage: Dict[str, Memory]

    def __init__(self):
        self._storage: Dict[str, Memory] = {}

    def get(self, key: str) -> Memory:
        value = self._storage.get(key)
        if value is None:
            raise ValueError(f"No memory found for key {key}")
        return TypeAdapter(Memory).validate_python(value)

    def get_all(self) -> List[Memory]:
        memories = []
        for value in self._storage.values():
            memories.append(TypeAdapter(Memory).validate_python(value))
        return memories

    def upsert(self, memory: Memory):
        self._storage[memory.key] = memory

    def close(self):
        self._storage = {}


class LocalMemoryStorage(InMemoryStorage):
    _local_storage_file = Path(__file__).resolve().parent.parent.parent / Path("memory_storage.json")

    def __init__(self):
        try:
            with open(self._local_storage_file, "rb") as file:
                self._storage = TypeAdapter(Dict[str, Memory]).validate_json(file.read())
        except FileNotFoundError:
            self._storage = {}

    def upsert(self, memory: Memory):
        super().upsert(memory)
        # TODO: Add debouncing to avoid writing to disk on every upsert
        with open(self._local_storage_file, "wb") as file:
            file.write(TypeAdapter(Dict[str, Memory]).dump_json(self._storage, indent=2))


class MemoryTool(StatefulTool):
    """
    Tool that manages memory.
    Can store memories.
    Can list the keys and titles of all the memories that are stored for later retrieval.
    """

    actions: tuple[type[Action], ...] = (GetMemoryTitlesAction, GetMemoryAction, UpsertMemoryAction)
    observations: tuple[type[Observation], ...] = (MemoryObservation, MemoryTitlesObservation)
    _action_map: dict[type[Action], Callable] = {}
    _memory_storage: MemoryStorage

    def model_post_init(self, __context: Any):
        self._memory_storage = LocalMemoryStorage()
        self._action_map = {
            GetMemoryTitlesAction: self.get_all_memories,
            GetMemoryAction: self.get_memory,
            UpsertMemoryAction: self.upsert_memory,
        }

    def get_all_memories(self, action: GetMemoryTitlesAction) -> MemoryTitlesObservation:
        memories = self._memory_storage.get_all()
        return MemoryTitlesObservation(titles=[MemoryTitle(key=memory.key, title=memory.title) for memory in memories])

    def get_memory(self, action: GetMemoryAction) -> MemoryObservation:
        memory = self._memory_storage.get(action.key)
        return MemoryObservation(key=memory.key, title=memory.title, data=memory.data)

    def upsert_memory(self, action: UpsertMemoryAction) -> MemoryObservation:
        self._memory_storage.upsert(Memory(key=action.key, title=action.title, data=action.data))
        return MemoryObservation(key=action.key, title=action.title, data=action.data)

    def execute_action(self, action: Action) -> MemoryObservation | MemoryTitlesObservation:
        action_type = type(action)
        if action_type in self._action_map:
            return self._action_map[action_type](action)
        raise ValueError(f"Unknown action: {action_type}")

    def close(self) -> None:
        self._memory_storage.close()



import json
import logging
from typing import Generator

from langchain_community.tools.tavily_search import TavilySearchResults

from examples.llama_user import UserModel, UserModelEvent, UserModelInstruction, UserModelTape
from tapeagents.core import MakeObservation, Prompt
from tapeagents.demo import Demo
from tapeagents.dialog_tape import (
    DialogContext,
    DialogTape,
    UserStep,
)
from tapeagents.environment import ToolEnvironment
from tapeagents.llms import LiteLLM, LLMOutput
from tapeagents.renderers import render_dialog_plain_text
from tapeagents.renderers.basic import BasicRenderer

from .annotator import GroundednessAnnotator
from .llama_user import USER_MODEL_TEMPLATE
from .openai_function_calling import FunctionCallingAgent

logging.basicConfig(level=logging.INFO)


class DemoUserModel(UserModel):
    def __init__(self, llm: LiteLLM, instruction: str):
        self.llm = llm
        super().__init__(instruction=instruction)

    def render_prompt(self, tape: UserModelTape) -> Prompt:
        (instruction_step,) = tape.steps
        assert isinstance(instruction_step, UserModelInstruction)
        conversation = render_dialog_plain_text(tape.context)
        assitant_message = USER_MODEL_TEMPLATE.format(
            conversation=conversation, user_model_prompt=instruction_step.instruction
        )
        return Prompt(messages=[{"role": "assistant", "content": assitant_message}])

    def generate_events(self, prompt: Prompt) -> Generator[UserModelEvent, None, None]:
        # say we don't need streaming for the agent model here
        for event in self.llm.generate(prompt):
            if m := event.output:
                assert isinstance(m, LLMOutput)
                try:
                    result = json.loads(m.content)
                except Exception:
                    raise ValueError(f"User model LLM returned invalid JSON: {event.output}")
                yield UserModelEvent(step=MakeObservation(new_observation=UserStep(**result)))
                return
        raise ValueError("User model LLM didn't return completion")


def user_model_demo():
    small_llm = LiteLLM(model_name="gpt-3.5-turbo")
    big_llm = LiteLLM(model_name="gpt-4-turbo")
    agent = FunctionCallingAgent(llms={"default": small_llm})
    environment = ToolEnvironment(tools=[TavilySearchResults()])
    init_dialog = DialogTape(context=DialogContext(tools=environment.get_tool_schemas()), steps=[])
    user_models = [
        DemoUserModel(big_llm, "ask about a celebrity"),
        DemoUserModel(big_llm, "ask about a movie"),
    ]
    demo = Demo(
        agent,
        init_dialog,
        environment,
        BasicRenderer(),
        user_models={um.instruction: um for um in user_models},
        annotator=GroundednessAnnotator(llms={"default": big_llm}),
    )
    demo.launch()


if __name__ == "__main__":
    user_model_demo()



from typing import Any

import yaml

from tapeagents.core import Action, Observation, Step, Thought
from tapeagents.dialog_tape import AssistantStep, DialogContext, SystemStep, UserStep
from tapeagents.environment import CodeExecutionResult, ExecuteCode
from tapeagents.renderers.basic import BasicRenderer
from tapeagents.tools.container_executor import CodeBlock
from tapeagents.view import Call, Respond


class PrettyRenderer(BasicRenderer):
    """Rendering enhancements for a handful of known steps."""

    def __init__(self, show_metadata=False, **kwargs):
        self.show_metadata = show_metadata
        super().__init__(**kwargs)

    @property
    def style(self):
        return super().style + (
            "<style>"
            ".observation { background-color: #baffc9;; }"
            ".error_observation { background-color: #dd0000; }"
            ".action { background-color: #cccccc; }"
            ".thought { background-color: #ffffdb; }"
            ".call { background-color: #ffffff; }"
            ".respond { background-color: #ffffff; }"
            ".step-header { margin: 2pt 2pt 2pt 0 !important; }"
            ".step-text { font-size: 12px; white-space: pre-wrap; word-wrap: break-word;}"
            "</style>"
        )

    def render_step(self, step: Step, index: int, **kwargs):
        title = type(step).__name__
        if isinstance(step, UserStep):
            role = "User"
            title = ""
            class_ = "observation"
        elif isinstance(step, SystemStep):
            role = "System"
            title = ""
            class_ = "observation"
        elif isinstance(step, AssistantStep):
            role = "Assistant"
            title = ""
            class_ = "action"
        elif isinstance(step, DialogContext):
            role = ""
            class_ = "observation"
        elif isinstance(step, Call):
            role = ""
            title = f"{step.metadata.agent.split('/')[-1]} calls {step.agent_name}"
            class_ = "call"
        elif isinstance(step, Respond):
            role = ""
            parts = step.metadata.agent.split("/")
            title = f"{parts[-1]} responds to {parts[-2]}" if len(parts) > 1 else f"{step.metadata.agent} responds"
            class_ = "respond"
        elif isinstance(step, Thought):
            role = "Thought"
            class_ = "thought"
        elif isinstance(step, Action):
            role = "Action"
            class_ = "action"
        elif isinstance(step, CodeExecutionResult):
            role = "Observation"
            class_ = "error_observation" if step.result.exit_code != 0 else "observation"
        elif isinstance(step, Observation):
            role = "Observation"
            class_ = "observation"
        else:
            raise ValueError(f"Unknown object type: {type(step)}")

        dump = step.model_dump()
        if not self.show_metadata:
            dump.pop("metadata", None)

        def pretty_yaml(d: dict):
            return yaml.dump(d, sort_keys=False, indent=2) if d else ""

        def maybe_fold(content: Any):
            content = str(content)
            summary = f"{len(content)} characters ..."
            if len(content) > 1000:
                return f"<details><summary>{summary}</summary>{content}</details>"
            return content

        if (content := getattr(step, "content", None)) is not None:
            # TODO: also show metadata here
            del dump["content"]
            text = pretty_yaml(dump) + ("\n" + maybe_fold(content) if content else "")
        elif isinstance(step, ExecuteCode):
            del dump["code"]

            def format_code_block(block: CodeBlock) -> str:
                return f"```{block.language}\n{block.code}\n```"

            code_blocks = "\n".join([format_code_block(block) for block in step.code])
            text = pretty_yaml(dump) + "\n" + maybe_fold(code_blocks)
        elif isinstance(step, CodeExecutionResult):
            del dump["result"]["output"]
            text = pretty_yaml(dump) + "\n" + maybe_fold(step.result.output)
        else:
            text = pretty_yaml(dump)

        index_str = f"[{index}]"
        header_text = title if not role else (role if not title else f"{role}: {title}")
        header = f"{index_str} {header_text}"

        return (
            f"<div class='basic-renderer-box {class_}'>"
            f"<h4 class='step-header'>{header}</h4>"
            f"<pre class='step-text'>{text}</pre>"
            f"</div>"
        )
